#Title:          FDRUtils.py (formerly PValue.py)
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
This script, based on PeptideProphet, computes the estimated probability that
a match is correct.  This probability is derived upon an F-Score.  The F-Score
for a match is a weighted sum of the length-corrected MQScore and the delta score.

We fit the distribution of F-scores as a mixture of two distributions:
A GAMMA DISTRIBUTION for false matches (with lower mean)
A NORMAL DISTRIBUTION for true matches (with higher mean)
Therefore, the probability that a match with a given F-Score is correct depends
upon the overall distribution of F-Scores for the rest of the run.

============================================================================
P-values for searches with shuffled-database:

Given a score cutoff, let the number of valid-protein hits above the
cutoff be V, and let the number of invalid-protein hits above the
cutoff be I.

# PVALUE WITH REMOVAL:
Throw out the I hits from invalid proteins.
Let TDB and FDB be the true and false database fractions.  (Note: For
a 1:1 database, TDB and FDB are both 0.5, so TDB/FDB equals 1.0)  Even
after throwing out hits from invalid proteins, there are still some
chance hits to true proteins.  The estimated number of hits from V
that are actually false is equal to: I*(TDB/FDB)

This, the odds that a match above this score cutoff is correct: 
(V - I*(TDB/FDB)) / V

This formulation of pvalue is the one normally used (e.g. in an
unmodified search).  Normally, we have no reason to keep any matches
to shuffled proteins; we only generate them in the first place so that
we can count them.

# PVALUE WITHOUT REMOVAL (-H command-line option):
Retain all hits.  As above, the number of
hits from V that are false is I*(TDB/FDB).  Thus, the odds that a
match above the score cutoff is correct:
(V - I*(TDB/FDB)) / (I+V)

"""
import os
import sys
import random
import math
import getopt
import traceback
import struct
import ResultsParser
import SelectProteins
import Learning
from Utils import *
Initialize()

try:
    from PIL import Image
    from PIL import ImageDraw
    from PIL import ImageFont
    # Fonts don't seem to work on Linux.  (Tried pdf, pcf, and pil formats...but no luck)
    # So, we'll content ourselves with a default font if we must:
    try:
        TheFont = ImageFont.truetype("Times.ttf", 12)
    except:
        TheFont = ImageFont.load_default()
except:
    print "(PIL not installed - image generation not available)"
    Image = None

class Colors:
    White = (255, 255, 255)
    Grey = (155, 155, 155)
    Background = (255, 255, 255)
    Black = (0, 0, 0)
    Green = (0, 155, 0)
    Red = (155, 0, 0)
    Blue = (0, 0, 155)

class Defaults:
    "Default F-score distribution; a starting point for E/M model fitting."
    MeanTrue = 4.48
    VarianceTrue = 1.50
    MeanFalse = 0.19
    VarianceFalse = 0.17
    PriorProbabilityTrue = 0.25
    GammaOffset = 0.3
    MQScoreWeight = 0.3
    DeltaScoreWeight = 1.5
    ###########################
    BlindMeanTrue = 5.0
    BlindVarianceTrue = 11.8
    BlindMeanFalse = -0.8
    BlindVarianceFalse = 0.7
    BlindPriorProbabilityTrue = 0.18
    BlindGammaOffset = 6.0
    BlindMQScoreWeight = 0.3
    BlindDeltaScoreWeight = 1.5

BLIND_MOD_PENALTY = 1.0
MIN_MQSCORE = -10.0

# Parse the scores from at most this many output files.  
MAX_RESULTS_FILES_TO_PARSE = 100

BIN_MULTIPLIER = 10.0
SQRT2PI = math.sqrt(2 * math.pi)

Cof = [76.18009172947146, -86.50532032941677,
    24.01409824083091, -1.231739572450155, 
    0.1208650973866179e-2, -0.5395239384952e-5]

def Gamma(Z):
    X = Z
    Y = Z
    Temp = X + 5.5
    Temp -= (X + 0.5) * math.log(Temp)
    Ser = 1.000000000190015
    for J in range(6):
        Y += 1
        Ser += Cof[J] / Y
    Z = -Temp + math.log(2.5066282746310005 * Ser / X)
    return math.exp(Z)

class Bag:
    pass

class PValueParser(ResultsParser.ResultsParser):
    def __init__(self):
        self.RetainBadMatches = 0
        self.LoadDistributionPath = None
        self.ScoreHistogram2 = {}
        self.ScoreHistogram3 = {}
        self.ShuffledScoreHistogram2 = {}
        self.ShuffledScoreHistogram3 = {}
        self.MinimumPeptideLength = 7
        self.VerboseFlag = 0
        self.GenerateImageFlag = 0
        self.MQScoreWeight = Defaults.MQScoreWeight
        self.DeltaScoreWeight = Defaults.DeltaScoreWeight
        self.GammaOffset = Defaults.GammaOffset
        self.BlindFlag = 0
        self.PValueCutoff = 0.1 # default
        # aminos -> location list
        self.PeptideDict = {}
        self.MaxDeltaScoreGap = -3.5
        self.DBPath = []
        self.PerformProteinSelection = 0
        self.ProteinPicker = None
        self.WriteTopMatchOnly = 0
        self.ShuffledDatabaseFraction = None
        self.RemoveShuffledMatches = 1
        # Overwrite existing files in -w target:
        self.OverwriteNewScoresFlag = 1
        self.ClusterInfoPath = None
        self.Columns = ResultsParser.Columns()
        ResultsParser.ResultsParser.__init__(self)
    def ReadDeltaScoreDistribution(self, FilePath):
        """
        Read delta-scores from a file, to compute the average delta-score.
        If passed a directory, iterate over all results files in the directory.
        """
        #
        self.AllSpectrumCount2 = 0
        self.AllSpectrumCount3 = 0
        self.MeanDeltaScore2 = 0
        self.MeanDeltaScore3 = 0
        self.ProcessResultsFiles(FilePath, self.ReadDeltaScoreDistributionFromFile, MAX_RESULTS_FILES_TO_PARSE)
        self.MeanDeltaScore2 /= max(1, self.AllSpectrumCount2)
        self.MeanDeltaScore3 /= max(1, self.AllSpectrumCount3)
        if self.VerboseFlag:
            print "Mean delta score ch1..2: %s over %s spectra"%(self.MeanDeltaScore2, self.AllSpectrumCount2)
            print "Mean delta score ch3: %s over %s spectra"%(self.MeanDeltaScore3, self.AllSpectrumCount3)
        if not self.MeanDeltaScore2:
            self.MeanDeltaScore2 = 0.001
        if not self.MeanDeltaScore3:
            self.MeanDeltaScore3 = 0.001
    def ReadDeltaScoreDistributionFromFile(self, FilePath):
        "Read delta-scores from a single file, to compute the average delta-score."
        print "Read delta-score distribution from %s..."%FilePath
        try:
            File = open(FilePath, "rb")
        except:
            traceback.print_exc()
            return
        OldSpectrum = None
        for FileLine in File.xreadlines():
            # Skip header lines and blank lines
            if FileLine[0] == "#":
                self.Columns.initializeHeaders(FileLine)
                continue
            if not FileLine.strip():
                continue
            Bits = list(FileLine.split("\t"))
            if len(Bits) <= self.Columns.getIndex("DeltaScore"):
                continue
            try:
                Charge = int(Bits[self.Columns.getIndex("Charge")])
                MQScore = float(Bits[self.Columns.getIndex("MQScore")])
                DeltaScore = float(Bits[self.Columns.getIndex("DeltaScoreOther")])
                Peptide = GetPeptideFromModdedName(Bits[self.Columns.getIndex("Annotation")])
                Spectrum = (os.path.basename(Bits[self.Columns.getIndex("SpectrumFile")]), Bits[self.Columns.getIndex("Scan#")])
            except:
                traceback.print_exc()
                print Bits
                continue # header line
            if Spectrum == OldSpectrum:
                continue
            
            OldSpectrum = Spectrum
            
            Length = len(Peptide.Aminos)
            if Length < self.MinimumPeptideLength:
                continue
            if DeltaScore < 0:
                print "## Warning: DeltaScore < 0!", Spectrum, FilePath
                print DeltaScore
                print MQScore
                print Bits
                raw_input()
                continue
            if Charge < 3:
                self.AllSpectrumCount2 += 1
                self.MeanDeltaScore2 += DeltaScore
                
            else:
                self.AllSpectrumCount3 += 1
                self.MeanDeltaScore3 += DeltaScore
        File.close()            
    def ReadScoreDistributionFromFile(self, FilePath):
        """
        Read F-scores from a single file, to compute the score histogram.
        """
        print "Read score distribution from %s..."%FilePath
        try:
            File = open(FilePath, "rb")
        except:
            traceback.print_exc()
            return
        OldSpectrum = None
        for FileLine in File.xreadlines():
            # Skip header lines and blank lines
            if FileLine[0] == "#":
                self.Columns.initializeHeaders(FileLine)
                continue
            if not FileLine.strip():
                continue            
            Bits = list(FileLine.split("\t"))
            try:
                Charge = int(Bits[self.Columns.getIndex("Charge")])
                MQScore = float(Bits[self.Columns.getIndex("MQScore")])
                DeltaScore = float(Bits[self.Columns.getIndex("DeltaScore")])
                Peptide = GetPeptideFromModdedName(Bits[self.Columns.getIndex("Annotation")])
                Protein = Bits[self.Columns.getIndex("ProteinName")]
                Spectrum = (Bits[self.Columns.getIndex("SpectrumFile")], Bits[self.Columns.getIndex("Scan#")])
            except:
                continue # header line
            if Spectrum == OldSpectrum:
                continue
            OldSpectrum = Spectrum
            Length = len(Peptide.Aminos)
            if Length < self.MinimumPeptideLength:
                continue
            if (Charge < 3):
                MeanDeltaScore = self.MeanDeltaScore2
            else:
                MeanDeltaScore = self.MeanDeltaScore3
            WeightedScore = self.MQScoreWeight * MQScore + self.DeltaScoreWeight * (DeltaScore / MeanDeltaScore)
            ScoreBin = int(round(WeightedScore * BIN_MULTIPLIER))
            Hit = 1
            if self.ClusterInfoPath:
                # Get this cluster's size:
                ClusterFileName = Bits[0].replace("/","\\").split("\\")[-1]
                ScanNumber = int(Bits[1])
                ClusterSize = self.ClusterSizes.get((ClusterFileName, ScanNumber), None)
                if not ClusterSize:
                    print "* Warning: ClusterSize not known for %s, %s"%(ClusterFileName, ScanNumber)
                else:
                    Hit = ClusterSize
            if Charge < 3:
                self.ScoreHistogram2[ScoreBin] = self.ScoreHistogram2.get(ScoreBin, 0) + Hit
            else:
                self.ScoreHistogram3[ScoreBin] = self.ScoreHistogram3.get(ScoreBin, 0) + Hit
            if self.ShuffledDatabaseFraction:
                if Protein[:3] == "XXX":
                    if Charge < 3:
                        self.ShuffledScoreHistogram2[ScoreBin] = self.ShuffledScoreHistogram2.get(ScoreBin, 0) + Hit
                    else:
                        self.ShuffledScoreHistogram3[ScoreBin] = self.ShuffledScoreHistogram3.get(ScoreBin, 0) + Hit
        File.close()
    def ProduceScoreDistributionImage(self, ImagePath, Charge3Flag = 0):
        """
        Write out, to the specified path, an image with f-score on the X-axis
        and p-value on the Y-axis.  If we fit a mixture model, plot the true
        and false (model) distributions; if we fit using shuffled proteins,
        plot the empirical distributions.
        """
        if Image == None:
            return
        
        if Charge3Flag:
            ScoreHistogram = self.ScoreHistogram3
            ShuffledScoreHistogram = self.ShuffledScoreHistogram3
        else:
            ScoreHistogram = self.ScoreHistogram2
            ShuffledScoreHistogram = self.ShuffledScoreHistogram2
        # Image size:
        self.Width = 900
        self.Height = 500
        self.LeftPadding = 50
        self.RightPadding = 80
        self.BottomPadding = 40
        self.TopPadding = 10
        self.PlotWidth = self.Width - (self.LeftPadding + self.RightPadding)
        self.PlotHeight = self.Height - (self.TopPadding + self.BottomPadding)
        self.PlotImage = Image.new("RGB", (self.Width, self.Height), Colors.Background)
        self.Draw = ImageDraw.Draw(self.PlotImage)
        # Find largest and smallest bins and entries:
        self.MaxScoreHistogramEntry = 10
        self.MaxScoreHistogramEntryValid = 10
        self.MaxScoreHistogramEntryInvalid = 10
        self.MaxScoreBin = -9999
        self.MinScoreBin = 9999
        self.TotalHistogramEntries = 0
        for (Bin, Entry) in ScoreHistogram.items():
            self.MaxScoreHistogramEntry = max(Entry, self.MaxScoreHistogramEntry)
            InvalidCount = ShuffledScoreHistogram.get(Bin, 0)
            ValidCount = max(0, Entry - InvalidCount)
            self.MaxScoreHistogramEntryValid = max(self.MaxScoreHistogramEntryValid, ValidCount)
            self.MaxScoreHistogramEntryInvalid = max(self.MaxScoreHistogramEntryInvalid, InvalidCount)
            self.MaxScoreBin = max(self.MaxScoreBin, Bin)
            self.MinScoreBin = min(self.MinScoreBin, Bin)
            self.TotalHistogramEntries += Entry
            #print "Bin %s: Valid %s, invalid %s"%(Bin, ValidCount, InvalidCount)
        self.BinCount = self.MaxScoreBin - self.MinScoreBin + 1
        # Draw the Y axis:
        self.Draw.line((self.LeftPadding, self.TopPadding, self.LeftPadding, self.TopPadding + self.PlotHeight), Colors.Black)
        self.Draw.line((self.Width - self.RightPadding, self.TopPadding, self.Width - self.RightPadding, self.TopPadding + self.PlotHeight), Colors.Black)
        Fraction = 0
        while Fraction <= 1.0:
            Y = self.TopPadding + self.PlotHeight * (1.0 - Fraction)
            Label = str(int(round(Fraction * self.MaxScoreHistogramEntry)))
            self.Draw.text((self.LeftPadding - 5 - len(Label)*5, Y - 6), Label, Colors.Black)
            self.Draw.line((self.LeftPadding - 5, Y, self.LeftPadding, Y), Colors.Black)
            Label = str(Fraction)
            self.Draw.text((self.Width - self.RightPadding + 10, Y - 6), Label, Colors.Black)
            self.Draw.line((self.Width - self.RightPadding, Y, self.Width - self.RightPadding + 5, Y), Colors.Black)
            Fraction += 0.1
        # Draw the X axis:
        self.Draw.line((self.LeftPadding, self.Height - self.BottomPadding, self.Width - self.RightPadding, self.Height - self.BottomPadding), Colors.Black)
        Bin = self.MinScoreBin
        while Bin % 10 != 0:
            Bin += 1
        while Bin < self.MaxScoreBin:
            BinNumber = Bin - self.MinScoreBin
            X = self.LeftPadding + BinNumber * self.PlotWidth / float(self.BinCount)            
            self.Draw.line((X, self.Height - self.BottomPadding - 2, X, self.Height - self.BottomPadding + 2), Colors.Black)
            Label = "%.1f"%(Bin / BIN_MULTIPLIER)
            self.Draw.text((X - len(Label) * 2.5, self.Height - self.BottomPadding + 2), Label, Colors.Black)
            Bin += 10
        if self.ShuffledDatabaseFraction != None:
            self.ProduceImageShuffledDB(Charge3Flag)
        else:
            self.ProduceImageMixtureModel(Charge3Flag)
        self.PlotImage.save(ImagePath)
        # Free:
        self.PlotImage = None
        self.Draw = None
    def ProduceImageShuffledDB(self, Charge3Flag = 0):
        if not Image:
            return

        if Charge3Flag:
            ScoreHistogram = self.ScoreHistogram3
            ShuffledScoreHistogram = self.ShuffledScoreHistogram3
            OddsTrue = self.OddsTrue3
        else:
            ScoreHistogram = self.ScoreHistogram2
            ShuffledScoreHistogram = self.ShuffledScoreHistogram2
            OddsTrue = self.OddsTrue2
        # Draw the legend:
        Y = self.Height - self.BottomPadding + 20
        self.Draw.line((105, Y, 125, Y), Colors.Black)
        self.Draw.rectangle((113, Y-2, 118, Y+2), Colors.Black)
        self.Draw.text((130, Y - 5), "p-value", Colors.Black)
        Y = self.Height - self.BottomPadding + 30
        self.Draw.line((105, Y, 125, Y), Colors.Blue)
        self.Draw.rectangle((113, Y-2, 118, Y+2), Colors.Grey)
        self.Draw.text((130, Y - 5), "All hits", Colors.Grey)
        Y = self.Height - self.BottomPadding + 20
        self.Draw.line((305, Y, 325, Y), Colors.Red)
        self.Draw.rectangle((313, Y-2, 318, Y+2), Colors.Green)
        self.Draw.text((330, Y - 5), "Valid proteins", Colors.Green)
        Y = self.Height - self.BottomPadding + 30
        self.Draw.line((305, Y, 325, Y), Colors.Red)
        self.Draw.rectangle((313, Y-2, 318, Y+2), Colors.Red)
        self.Draw.text((330, Y - 5), "Invalid proteins", Colors.Red)
        # Loop over bins, plotting distributions:
        PrevYOdds = None
        PrevYAll = None
        PrevYTrue = None
        PrevYFalse = None
        PrevX = None
        for Bin in range(self.MinScoreBin, self.MaxScoreBin + 1):
            BinNumber = Bin - self.MinScoreBin
            XX = self.LeftPadding + (BinNumber * self.PlotWidth / float(self.BinCount))
            # p-value:
            PValue = 1.0 - OddsTrue[Bin]
            YOdds = self.Height - self.BottomPadding - self.PlotHeight * PValue
            self.Draw.rectangle((XX - 2, YOdds - 2, XX + 2, YOdds + 2), Colors.Black)
            if PrevYOdds != None:
                self.Draw.line((PrevX, PrevYOdds, XX, YOdds), Colors.Black)
            # Overall:
            Count = ScoreHistogram.get(Bin, 0)
            YAll = self.Height - self.BottomPadding - self.PlotHeight * Count / float(self.MaxScoreHistogramEntry)
            self.Draw.rectangle((XX - 2, YAll - 2, XX + 2, YAll + 2), Colors.Grey)
            if (PrevYAll):
                self.Draw.line((PrevX, PrevYAll, XX, YAll), Colors.Grey)
            # Invalid:
            CountInvalid = ShuffledScoreHistogram.get(Bin, 0)
            YFalse = self.Height - self.BottomPadding - self.PlotHeight * CountInvalid / float(self.MaxScoreHistogramEntryInvalid)
            self.Draw.rectangle((XX - 2, YFalse - 2, XX + 2, YFalse + 2), Colors.Red)
            if (PrevYFalse):
                self.Draw.line((PrevX, PrevYFalse, XX, YFalse), Colors.Red)
            # Valid:
            CountValid = Count - CountInvalid
            YTrue = self.Height - self.BottomPadding - self.PlotHeight * CountValid / float(self.MaxScoreHistogramEntryValid)
            self.Draw.rectangle((XX - 2, YTrue - 2, XX + 2, YTrue + 2), Colors.Green)
            #print "Bin %s: Valid %s/%s invalid %s/%s"%(Bin, CountValid, self.MaxScoreHistogramEntryValid, CountInvalid, self.MaxScoreHistogramEntryInvalid)
            if (PrevYTrue):
                self.Draw.line((PrevX, PrevYTrue, XX, YTrue), Colors.Green)
            # Remember these values, for linking to the next in the series:
            PrevX = XX
            PrevYOdds = YOdds
            PrevYAll = YAll
            PrevYFalse = YFalse
            PrevYTrue = YTrue
    def ProduceImageMixtureModel(self, Charge3Flag = 0):
        """
        Helper for ProduceScoreDistributionImage, if we're using a mixture
        model (not a shuffled database)
        """
        if Image == None:
            return
        if Charge3Flag:
            ScoreHistogram = self.ScoreHistogram3
            ShuffledScoreHistogram = self.ShuffledScoreHistogram3
            MixtureModel = self.MixtureModel3
            OddsTrue = self.OddsTrue3
        else:
            ScoreHistogram = self.ScoreHistogram2
            ShuffledScoreHistogram = self.ShuffledScoreHistogram2
            MixtureModel = self.MixtureModel2
            OddsTrue = self.OddsTrue2
        # Draw the legend:
        Y = self.Height - self.BottomPadding + 20
        self.Draw.line((105, Y, 125, Y), Colors.Black)
        self.Draw.rectangle((113, Y-2, 118, Y+2), Colors.Black)
        self.Draw.text((130, Y - 5), "Empirical score distribution", Colors.Black)
        Y = self.Height - self.BottomPadding + 30
        self.Draw.line((105, Y, 125, Y), Colors.Blue)
        self.Draw.rectangle((113, Y-2, 118, Y+2), Colors.Blue)
        self.Draw.text((130, Y - 5), "Probability true (1-pvalue)", Colors.Blue)
        Y = self.Height - self.BottomPadding + 20
        self.Draw.line((305, Y, 325, Y), Colors.Red)
        self.Draw.rectangle((313, Y-2, 318, Y+2), Colors.Red)
        self.Draw.text((330, Y - 5), "Gamma dist. (fit to false matches)", Colors.Red)
        Y = self.Height - self.BottomPadding + 30
        self.Draw.line((305, Y, 325, Y), Colors.Green)
        self.Draw.rectangle((313, Y-2, 318, Y+2), Colors.Green)
        self.Draw.text((330, Y - 5), "Normal dist. (fit to true matches)", Colors.Green)
        Y = self.Height - self.BottomPadding + 20
        self.Draw.line((555, Y, 575, Y), Colors.Grey)
        self.Draw.rectangle((563, Y-2, 568, Y+2), Colors.Grey)
        self.Draw.text((580, Y - 5), "Fitted mixture model", Colors.Grey)
        # Draw the plot of OBSERVED SCORES:
        PrevX = None
        PrevY = None
        for Bin in range(self.MinScoreBin, self.MaxScoreBin + 1):
            BinNumber = Bin - self.MinScoreBin
            X = self.LeftPadding + BinNumber * self.PlotWidth / float(self.BinCount)
            Count = ScoreHistogram.get(Bin, 0)
            Y = self.Height - self.BottomPadding - self.PlotHeight * Count / float(self.MaxScoreHistogramEntry)
            self.Draw.rectangle((X - 2, Y - 2, X + 2, Y + 2), Colors.Black)
            if PrevX != None:
                self.Draw.line((PrevX, PrevY, X, Y), Colors.Black)
            PrevX = X
            PrevY = Y
        #######################################################
        # Find the scaling factor for the MERGED distribution:
        ComboDistTotal = 0
        for Bin in range(self.MinScoreBin, self.MaxScoreBin + 1):
            TrueScore = Bin / BIN_MULTIPLIER
            Pow = - ((TrueScore - MixtureModel.MeanTrue)**2) / (2 * MixtureModel.VarianceTrue)
            TrueNormal = math.exp(Pow) / (MixtureModel.StdDevTrue * SQRT2PI)
            GX = max(0.01, TrueScore + MixtureModel.GammaOffset)
            FalseGamma = math.pow(GX, MixtureModel.KFalse - 1) * math.exp(-GX / MixtureModel.ThetaFalse) / MixtureModel.GammaDemonFalse
            ComboDist = TrueNormal * MixtureModel.PriorProbabilityTrue + (1.0 - MixtureModel.PriorProbabilityTrue) * FalseGamma
            ComboDistTotal += ComboDist
        YFittedScalingFactor = self.TotalHistogramEntries / ComboDistTotal
        #######################################################
        # Draw the plot of the FALSE HIT GAMMA and TRUE HIT NORMAL and MERGED distributions:
        PrevX = None
        PrevYNormal = None
        PrevYGamma = None
        PrevYOdds = None
        PrevYFitted = None
        for Bin in range(self.MinScoreBin, self.MaxScoreBin + 1):
            BinNumber = Bin - self.MinScoreBin
            XX = self.LeftPadding + (BinNumber * self.PlotWidth / float(self.BinCount))
            TrueScore = Bin / BIN_MULTIPLIER
            Pow = - ((TrueScore - MixtureModel.MeanTrue)**2) / (2 * MixtureModel.VarianceTrue)
            TrueNormal = math.exp(Pow) / (MixtureModel.StdDevTrue * SQRT2PI)
            GX = max(0.01, TrueScore + MixtureModel.GammaOffset)
            FalseGamma = math.pow(GX, MixtureModel.KFalse - 1) * math.exp(-GX / MixtureModel.ThetaFalse) / MixtureModel.GammaDemonFalse
            YNormal = self.Height - self.BottomPadding - self.PlotHeight * TrueNormal
            # Normal distribution:
            self.Draw.rectangle((XX - 2, YNormal - 2, XX + 2, YNormal + 2), Colors.Green)
            if PrevX != None:
                self.Draw.line((PrevX, PrevYNormal, XX, YNormal), Colors.Green)
            # Gamma distribution:
            YGamma = self.Height - self.BottomPadding - self.PlotHeight * FalseGamma
            self.Draw.rectangle((XX - 2, YGamma - 2, XX + 2, YGamma + 2), Colors.Red)
            if PrevX != None:
                self.Draw.line((PrevX, PrevYGamma, XX, YGamma), Colors.Red)
            # Fitted curve:
            ComboDist = TrueNormal * MixtureModel.PriorProbabilityTrue + (1.0 - MixtureModel.PriorProbabilityTrue) * FalseGamma
            YFitted = ComboDist * YFittedScalingFactor / self.MaxScoreHistogramEntry
            YFitted = self.Height - self.BottomPadding - YFitted * self.PlotHeight
            #print TrueNormal, FalseGamma, self.AllSpectrumCount, ComboDist, YFitted
            self.Draw.rectangle((XX - 2, YFitted - 2, XX + 2, YFitted + 2), Colors.Grey)
            if PrevX != None:
                self.Draw.line((PrevX, PrevYFitted, XX, YFitted), Colors.Grey)
            # P-Value:
            PValue = 1.0 - OddsTrue.get(Bin, 0)
            YOdds = self.Height - self.BottomPadding - self.PlotHeight * PValue
            self.Draw.rectangle((XX - 2, YOdds - 2, XX + 2, YOdds + 2), Colors.Blue)
            if PrevX != None:
                self.Draw.line((PrevX, PrevYOdds, XX, YOdds), Colors.Blue)
            # Remember these points' coords for drawing lines next time:
            PrevX = XX
            PrevYNormal = YNormal
            PrevYGamma = YGamma
            PrevYOdds = YOdds
            PrevYFitted = YFitted
    def FitMixtureModel(self):
        self.MixtureModel2 = Learning.MixtureModelClass()
        self.MixtureModel2.Model(None, self.ScoreHistogram2)
        self.OddsTrue2 = self.MixtureModel2.OddsTrue
        self.MixtureModel3 = Learning.MixtureModelClass()
        self.MixtureModel3.Model(None, self.ScoreHistogram3)
        self.OddsTrue3 = self.MixtureModel3.OddsTrue        
        return 1
    def SavePValueDistribution(self, Charge3Flag = 0):
        """
        Write out a p-value distribution derived from forward+shuffled database
        """
        if Charge3Flag:
            OddsTrue = self.OddsTrue3
            MeanDeltaScore = self.MeanDeltaScore3
            ScoreHistogram = self.ScoreHistogram3
            ShuffledScoreHistogram = self.ShuffledScoreHistogram3
        else:
            OddsTrue = self.OddsTrue2
            MeanDeltaScore = self.MeanDeltaScore2
            ScoreHistogram = self.ScoreHistogram2
            ShuffledScoreHistogram = self.ShuffledScoreHistogram2
        Keys = OddsTrue.keys()
        if not Keys:
            return
        MinBin = min(Keys)
        MaxBin = max(Keys)
        self.OutputDistributionFile.write("#MeanDeltaScore\t%s\n"%MeanDeltaScore)
        self.OutputDistributionFile.write("#BlindFlag\t%s\n"%self.BlindFlag)
        if self.ShuffledDatabaseFraction != None:
            Header = "#Bin\tFDR\tTotalHits\tHitsValid\tHitsInvalid\tPeptideFDR\tPeptidesValid\tPeptidesInvalid\tProteinFDR\tProteinsValid\tProteinsInvalid\t\n"
        else:
            Header = "#Bin\tFDR\tTotalHits\t\n"
        self.OutputDistributionFile.write(Header)
        if self.ShuffledDatabaseFraction != None:
            # Count the total number of true hits, false hits, true peptides, false peptides...
            CumulativeTrueHits = 0
            CumulativeFalseHits = 0
            for Bin in range(MinBin, MaxBin + 1):
                AllHits = ScoreHistogram.get(Bin, 0)
                FalseHits = ShuffledScoreHistogram.get(Bin, 0)
                CumulativeFalseHits += FalseHits
                CumulativeTrueHits += (AllHits - FalseHits)
            if self.ProteinPicker:
                ######################################################
                # Peptides:
                ValidPeptides = {}
                InvalidPeptides = {}
                CumulativeTruePeptides = 0
                CumulativeFalsePeptides = 0
                BestScoreByProtein = {}
                for (Peptide, Score) in self.ProteinPicker.BestScoresByPeptide.items():
                    Bin = int(round(Score / BIN_MULTIPLIER))
                    ProteinID = self.ProteinPicker.PeptideProteins.get(Peptide, None)
                    if not ProteinID:
                        print "*** Warning: Peptide '%s' was never assigned to a protein!"%Peptide
                        LocationList = self.ProteinPicker.PeptideDict[Peptide]
                        print LocationList
                        for (ProteinID, Pos) in LocationList:
                            print ProteinID, self.ProteinPicker.ProteinNames[ProteinID], self.ProteinPicker.ProteinPeptideCounts[ProteinID]
                        continue # shouldn't occur!
                    ProteinName = self.ProteinPicker.ProteinNames[ProteinID]
                    OldScore = BestScoreByProtein.get(ProteinID, -9999)
                    BestScoreByProtein[ProteinID] = max(OldScore, Score)
                    if ProteinName[:3] == "XXX":
                        InvalidPeptides[Bin] = InvalidPeptides.get(Bin, 0) + 1
                        CumulativeFalsePeptides += 1
                    else:
                        ValidPeptides[Bin] = ValidPeptides.get(Bin, 0) + 1
                        CumulativeTruePeptides += 1
                ######################################################
                # Proteins:
                ValidProteins = {}
                InvalidProteins = {}
                CumulativeTrueProteins = 0
                CumulativeFalseProteins = 0
                for (ProteinID, Score) in BestScoreByProtein.items():
                    ProteinName = self.ProteinPicker.ProteinNames[ProteinID]
                    if ProteinName[:3] == "XXX":
                        InvalidProteins[Bin] = InvalidProteins.get(Bin, 0) + 1
                        CumulativeFalseProteins += 1
                    else:
                        ValidProteins[Bin] = ValidProteins.get(Bin, 0) + 1
                        CumulativeTrueProteins += 1
        for Bin in range(MinBin, MaxBin + 1):
            FDR = 1.0 - OddsTrue[Bin]
            AllHits = ScoreHistogram.get(Bin, 0)
            self.OutputDistributionFile.write("%s\t%s\t%s\t"%(Bin, FDR, AllHits))
            if self.ShuffledDatabaseFraction != None:
                FalseHits = ShuffledScoreHistogram.get(Bin, 0)
                TrueHits = AllHits - FalseHits
                CumulativeTrueHits -= TrueHits
                CumulativeFalseHits -= FalseHits                
                self.OutputDistributionFile.write("%s\t%s\t"%(CumulativeTrueHits, CumulativeFalseHits))
                if self.ProteinPicker:
                    # Peptide FDR:
                    FalseWithinTrue = min(CumulativeTruePeptides, CumulativeFalsePeptides * self.ShuffledScalingFactor)
                    PeptideFDR = FalseWithinTrue / float(max(1, CumulativeTruePeptides))
                    self.OutputDistributionFile.write("%.4f\t%s\t%s\t"%(PeptideFDR, CumulativeTruePeptides, CumulativeFalsePeptides))
                    CumulativeTruePeptides -= ValidPeptides.get(Bin, 0)
                    CumulativeFalsePeptides -= InvalidPeptides.get(Bin, 0)
                    # Protein FDR:
                    FalseWithinTrue = min(CumulativeTrueProteins, CumulativeFalseProteins * self.ShuffledScalingFactor)
                    ProteinFDR = FalseWithinTrue / float(max(1, CumulativeTrueProteins))
                    self.OutputDistributionFile.write("%.4f\t%s\t%s\t"%(ProteinFDR, CumulativeTrueProteins, CumulativeFalseProteins))
                    CumulativeTrueProteins -= ValidProteins.get(Bin, 0)
                    CumulativeFalseProteins -= InvalidProteins.get(Bin, 0)
            self.OutputDistributionFile.write("\n")
    def LoadPValueDistribution(self, FileName):
        Charge3Flag = 0
        
        File = open(FileName, "rb")

        for FileLine in File.xreadlines():
            Bits = list(FileLine.strip().split("\t"))
            if len(Bits) < 2:
                continue            
            if FileLine[0] == "#":
                # Header line.  Parse special lines:
                Name = Bits[0][1:]
                if Name == "BlindFlag":
                    self.BlindFlag = int(Bits[1])
                elif Name == "MeanDeltaScore":
                    if Charge3Flag:
                        self.MeanDeltaScore3 = float(Bits[1])
                        OddsTrue = {}
                        self.OddsTrue3 = OddsTrue
                    else:
                        self.MeanDeltaScore2 = float(Bits[1])
                        OddsTrue = {}
                        self.OddsTrue2 = OddsTrue
                        
                else:
                    print "(Skipping comment '%s', not understood)"%Bits[0]
                continue
            Bin = int(Bits[0])
            OddsTrue[Bin] = 1.0 - float(Bits[1])
            Charge3Flag = 1 #We've gotten past all the comments for charges 1 and 2, so the next time
            #we see a '#' it will be for charge 3
        File.close()
        if self.BlindFlag:
            self.MQScoreWeight = Defaults.BlindMQScoreWeight
            self.DeltaScoreWeight = Defaults.BlindDeltaScoreWeight
            self.GammaOffset = Defaults.BlindGammaOffset
        else:
            self.MQScoreWeight = Defaults.MQScoreWeight
            self.DeltaScoreWeight = Defaults.DeltaScoreWeight
            self.GammaOffset = Defaults.GammaOffset
    def WriteMatchesForSpectrum(self, MatchesForSpectrum, OutFile):
        if self.WriteTopMatchOnly:
            MatchesForSpectrum = MatchesForSpectrum[0:1]
        for Match in MatchesForSpectrum:
            # If we have a shuffled database (-S option), then by default we get to throw shuffled-protein
            # matches away for free.  We don't get to keep runners-up to them, though!!
            if Match.ProteinName[:3] == "XXX" and self.ShuffledDatabaseFraction != None and self.RemoveShuffledMatches:
                break
            # Skip matches with poor delta-score:
            if Match.DeltaScore < self.MaxDeltaScoreGap and not self.RetainBadMatches:
                continue
            # Skip short matches:
            Length = len(Match.Peptide.Aminos)
            if Length < self.MinimumPeptideLength:
                continue
            if Match.Charge < 3:
                MeanDeltaScore = self.MeanDeltaScore2
            else:
                MeanDeltaScore = self.MeanDeltaScore3
            WeightedScore = self.MQScoreWeight * Match.MQScore + self.DeltaScoreWeight * (Match.DeltaScore / MeanDeltaScore)
            ScoreBin = int(round(WeightedScore * BIN_MULTIPLIER))
            if Match.Charge < 3:
                TrueOdds = self.OddsTrue2.get(ScoreBin, None)
            else:
                TrueOdds = self.OddsTrue3.get(ScoreBin, None)
            if TrueOdds == None:
                if ScoreBin < 0:
                    TrueOdds = 0.00001
                else:
                    TrueOdds = 0.99999
            else:
                TrueOdds = max(0.00001, min(TrueOdds, 0.99999))
            Match.PValue = (1.0 - TrueOdds)
            Match.Bits[self.Columns.getIndex("F-Score")] = "%s"%WeightedScore
            Match.Bits[self.Columns.getIndex("InspectFDR")] = "%s"%Match.PValue
            if self.ProteinPicker:
                # Replace the original protein with the "correct" one:
                ProteinID = self.ProteinPicker.PeptideProteins.get(Match.Peptide.Aminos, None)
                if ProteinID != None:
                    Match.Bits[self.Columns.getIndex("RecordNumber")] = str(ProteinID)
                    Match.Bits[self.Columns.getIndex("Protein")] = self.ProteinPicker.ProteinNames[ProteinID]
            if (not self.RetainBadMatches):
                if (Match.PValue > self.PValueCutoff):
                    continue
                # Sometimes things with a horrible MQScore get a good pvalue.
                # We want to exclude these.
                if Match.MQScore < MIN_MQSCORE:
                    continue
            self.LinesAcceptedCount += 1
            OutFile.write(string.join(Match.Bits, "\t"))
            OutFile.write("\n")
    def WriteFixedScores(self, OutputPath):
        self.TotalLinesAcceptedCount = 0
        self.TotalLinesSecondPass = 0
        self.WriteScoresPath = OutputPath
        # Make the output directory, if it doesn't exist already.
        # Assume: OutputPath is a directory if ReadScoresPath is a directory,
        # and OutputPath is a file if ReadScoresPath is a file.
        if os.path.isdir(self.ReadScoresPath):
            DirName = OutputPath
        else:
            DirName = os.path.split(OutputPath)[0]
        try:
            os.makedirs(DirName)
        except:
            pass
        self.ProcessResultsFiles(self.ReadScoresPath, self.WriteFixedScoresFile)
        print "Total accepted lines: %s of %s"%(self.TotalLinesAcceptedCount, self.TotalLinesSecondPass)
    def WriteFixedScoresFile(self, Path):
        if os.path.isdir(self.ReadScoresPath):
            OutputPath = os.path.join(self.WriteScoresPath, os.path.split(Path)[1])
        else:
            OutputPath = self.WriteScoresPath
        if (not self.OverwriteNewScoresFlag) and os.path.exists(OutputPath):
            return
        try:
            InFile = open(Path, "rb")
            OutFile = open(OutputPath, "wb")
            LineCount = 0
            self.LinesAcceptedCount = 0
            OldSpectrum = None
            MatchesForSpectrum = []
            for FileLine in InFile.xreadlines():
                # Lines starting with # are comments (e.g. header line), and are written out as-is:
                if FileLine[0] == "#":
                    self.Columns.initializeHeaders(FileLine)
                    OutFile.write(FileLine)
                    continue
                Bits = list(FileLine.strip().split("\t"))
                Match = Bag()
                try:
                    Match.Bits = Bits
                    Match.Charge = int(Bits[self.Columns.getIndex("Charge")])
                    Match.MQScore = float(Bits[self.Columns.getIndex("MQScore")])
                    #Match.DeltaScoreAny = float(Bits[self.Columns.DeltaScoreAny])
                    Match.DeltaScore = float(Bits[self.Columns.getIndex("DeltaScore")])
                    Match.Peptide = GetPeptideFromModdedName(Bits[self.Columns.getIndex("Annotation")])
                    Match.ProteinName = Bits[self.Columns.getIndex("Protein")]
                except:
                    continue
                LineCount += 1
                Spectrum = (Bits[0], Bits[1])
                if Spectrum != OldSpectrum:
                    self.WriteMatchesForSpectrum(MatchesForSpectrum, OutFile)
                    MatchesForSpectrum = []
                OldSpectrum = Spectrum
                MatchesForSpectrum.append(Match)
            # Finish the last spectrum:
            self.WriteMatchesForSpectrum(MatchesForSpectrum, OutFile)
            InFile.close()
            OutFile.close()
            print "%s\t%s\t%s\t"%(Path, LineCount, self.LinesAcceptedCount)
            self.TotalLinesAcceptedCount += self.LinesAcceptedCount
            self.TotalLinesSecondPass += LineCount
        except:
            traceback.print_exc()
            print "* Error filtering annotations from '%s' to '%s'"%(Path, OutputPath)
    def ComputePValuesWithShuffled(self, Charge3Flag = 0):
        """
        Set self.OddsTrue using results from a partially-shuffled database.
        Given a score cutoff we assume that, above the score cutoff, there are
        T hits from valid proteins and F hits from invalid proteins.
        
        # PVALUE WITH REMOVAL:
        Let TDB and FDB be the true and false database fractions (FDB = self.ShuffledDatabaseFraction)
        After filtering out all F hits from invalid proteins, there are still some
        chance hits to true proteins.  The estimated number of hits from T that are actually
        false is equal to F*(TDB/FDB).  This, the odds true for this cutoff is:
        1.0 - (F*(TDB/FDB) / T)
        
        # PVALUE WITHOUT REMOVAL:
        The odds true for the cutoff is simply T/(T+F).  
        """
        OddsTrue = {}
        if Charge3Flag:
            self.OddsTrue3 = OddsTrue
            ScoreHistogram = self.ScoreHistogram3
            ShuffledScoreHistogram = self.ShuffledScoreHistogram3
        else:
            self.OddsTrue2 = OddsTrue
            ScoreHistogram = self.ScoreHistogram2
            ShuffledScoreHistogram = self.ShuffledScoreHistogram2
        CumulativeHits = 0
        CumulativeHitsTrue = 0
        CumulativeHitsFalse = 0
        Keys = ScoreHistogram.keys()
        Keys.sort()
        if not Keys:
            # There are NO HITS for this charge state.
            return
        MinKey = Keys[0]
        MaxKey = Keys[-1]
        TrueFraction = 1.0 - self.ShuffledDatabaseFraction
        self.ShuffledScalingFactor = TrueFraction / self.ShuffledDatabaseFraction
        for Key in range(MaxKey, MinKey - 1, -1):
            AllHits = ScoreHistogram.get(Key, 0)
            FalseHits = ShuffledScoreHistogram.get(Key, 0)
            ValidHits = AllHits - FalseHits
            CumulativeHitsTrue += ValidHits
            CumulativeHitsFalse += FalseHits
            FalseWithinTrue = min(CumulativeHitsTrue, CumulativeHitsFalse * self.ShuffledScalingFactor)
            ##NEC_MOD
            #if FalseWithinTrue == 0:
            #    FalseWithinTrue = 1
            if self.RemoveShuffledMatches:
                # OddsTrue = (V - I*(TDB/FDB)) / V
                BinOddsTrue = max(0, CumulativeHitsTrue - FalseWithinTrue) / float(max(1, CumulativeHitsTrue))
                
            else:
                # OddsTrue = (V - I*(TDB/FDB)) / (I+V)
                BinOddsTrue = max(0, CumulativeHitsTrue - FalseWithinTrue) / float(max(1, CumulativeHitsTrue + CumulativeHitsFalse))
            if self.VerboseFlag:
                # Bin, true, false, cumtrue, cumfalse
                Str = "%s\t%s\t%s\t%s\t%s\t"%(Key, ValidHits, FalseHits, CumulativeHitsTrue, CumulativeHitsFalse)
                Str += "%.5f\t%.5f\t"%(BinOddsTrue, 1.0 - BinOddsTrue)
                print Str
            OddsTrue[Key] = BinOddsTrue
        if self.VerboseFlag:
            print "\n\n"
    def SelectProteins(self, PValueCutoff, ReadScoresPath):
        """
        Using SelectProteins, assign each peptide the the most reasonable "owner" protein.
        """
        # Select the F-score cutoff:
        FScoreCutoff2 = 9999
        FScoreCutoff3 = 9999
        for FScoreBin in self.OddsTrue2.keys():
            OddsTrue = self.OddsTrue2[FScoreBin]
            if (1.0 - OddsTrue) <= PValueCutoff:
                if (FScoreBin / BIN_MULTIPLIER) < FScoreCutoff2:
                    FScoreCutoff2 = FScoreBin / BIN_MULTIPLIER
        for FScoreBin in self.OddsTrue3.keys():
            OddsTrue = self.OddsTrue3[FScoreBin]
            if (1.0 - OddsTrue) <= PValueCutoff:
                if (FScoreBin / BIN_MULTIPLIER) < FScoreCutoff3:
                    FScoreCutoff3 = FScoreBin / BIN_MULTIPLIER
        self.ProteinPicker.FScoreCutoff2 = FScoreCutoff2
        self.ProteinPicker.FScoreCutoff3 = FScoreCutoff3
        self.ProteinPicker.MeanDeltaScore2 = self.MeanDeltaScore2
        self.ProteinPicker.MeanDeltaScore3 = self.MeanDeltaScore3
        self.ProcessResultsFiles(ReadScoresPath, self.ProteinPicker.ParseAnnotations)
        # We've COUNTED the protein hits.  Now ask the picker to decide which
        # protein 'owns' each peptide:
        self.ProteinPicker.ChooseProteins()
    def SetOutputDistributionPath(self, Path):
        self.OutputDistributionPath = Path
        self.OutputDistributionFile = open(Path, "wb")
    def ParseClusterInfo(self):
        """
        Parse cluster-sizes from an info file.  
        """
        self.ClusterSizes = {}
        File = open(self.ClusterInfoPath, "rb")
        for FileLine in File.xreadlines():
            Bits = FileLine.split()
            try:
                ScanNumber = int(Bits[1])
                ClusterSize = int(Bits[2])
            except:
                print "* Skipping this line:", FileLine
            self.ClusterSizes[(Bits[0], ScanNumber)] = ClusterSize
    def ParseCommandLine(self, Arguments):
        (Options, Args) = getopt.getopt(Arguments, "l:s:r:w:m:bp:vixzd:a1S:HX:")
        OptionsSeen = {}
        self.SaveDistributionPath = "PValues.txt" # default
        self.ReadScoresPath = None
        self.WriteScoresPath = None
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-l":
                if not os.path.exists(Value):
                    print "** Error: can't read p-value distribution from file '%s'"%Value
                    return 0
                self.LoadDistributionPath = Value
            elif Option == "-p":
                self.PValueCutoff = float(Value)
            elif Option == "-s":
                self.SaveDistributionPath = Value
            elif Option == "-x":
                self.RetainBadMatches = 1
            elif Option == "-b":
                self.BlindFlag = 1
                self.MQScoreWeight = Defaults.BlindMQScoreWeight
                self.DeltaScoreWeight = Defaults.BlindDeltaScoreWeight
                self.GammaOffset = Defaults.BlindGammaOffset
            elif Option == "-r":
                self.ReadScoresPath  = Value
            elif Option == "-w":
                self.WriteScoresPath = Value
            elif Option == "-m":
                MAX_RESULTS_FILES_TO_PARSE = int(Value)
            elif Option == "-v":
                self.VerboseFlag = 1
            elif Option == "-i":
                self.GenerateImageFlag = 1
            elif Option == "-d":
                if not os.path.exists(Value):
                    print "** Error: couldn't find database file '%s'\n\n"%Value
                    print UsageInfo
                    sys.exit(1)
                self.DBPath.append(Value)
            elif Option == "-a":
                self.PerformProteinSelection = 1
            elif Option == "-1":
                self.WriteTopMatchOnly = 1
            elif Option == "-S":
                self.ShuffledDatabaseFraction = float(Value)
                if self.ShuffledDatabaseFraction <= 0 or self.ShuffledDatabaseFraction >= 1:
                    print "* Invalid value for -S: %s"%Value
                    return 0
            elif Option == "-H":
                self.RemoveShuffledMatches = 0
            elif Option == "-X":
                # Undocumented option for CLUSTER searches:
                self.ClusterInfoPath = Value
                self.ParseClusterInfo()
            else:
                print "** Unknown option:", Option, Value
        # Check validity of options:
        if self.PerformProteinSelection and not self.DBPath:
            print "* Error: -a option requires -d option!"
            return 0
        # No major problems - return TRUE for success.
        return 1

UsageInfo = """
FDRUtils.py - Compute probability that each match from a tandem MS
peptide database search is correct.  Write out an updated results file containing
only the high-quality results.

Parameters:
 -r [FILENAME] Read results from filename (and fit the probability mixture
    model to these results).  If the option value is a directory, we'll read
    all the results-files from the directory.
 -w [FILENAME] Write re-scored results to a file.
 -l [FILENAME] Load p-value distribution from a file (written out earlier
    with -s option)

Protein selection can be performed, replacing the protein identification
with a parsimonious set of protein IDs (using a simple iterative
approach).  The following options are required for protein selection:
 -a: Replace protein identifications with a "parsimonious" set of protein IDs.
     Requires -d option!
 -d [FILENAME] Database (.trie file) searched
 -S [FRACTION]: (see below)

Other options:
 -S [FRACTION]: The fraction of the database consisting of shuffled
    proteins.  For instance, if you use a 1:1 mix of valid and invalid
    proteins, use -S 0.5.  If this option is set, p-values will be set using
    the number of matches to shuffled proteins, whose names begin with XXX
 -s [FILENAME] Save p-value distribution to a file.
 -i Write a .png image of the distribution graph (requires PIL)
 -p [NUM] FDR cutoff for saving results; by default, 0.1
 -b Blind search (use different score/deltascore weighting)
 -x If the -x flag is passed, even "bad" matches are written out (no p-value
    filtering is performed)
 -1 Write only the top hit for each spectrum, even if "good" runners-up exist

Internal use only:
 -v Verbose output (for debugging)
 -H Retain matches to shuffled proteins.  Used for further processing ONLY.
    
Example:
  FDRUtils.py -r ShewanellaResults -s ShewFDR.txt -w ShewanellaFiltered
     -p 0.05 -d database\Shew.trie -a 
"""

def Main(Parser = None):
    global MAX_RESULTS_FILES_TO_PARSE
    
    if not Parser:
        Parser = PValueParser()
        Result = Parser.ParseCommandLine(sys.argv[1:])
        if not Result:
            print UsageInfo
            return
    if Parser.DBPath and Parser.PerformProteinSelection:
        Parser.ProteinPicker = SelectProteins.ProteinSelector()
        Parser.ProteinPicker.LoadMultipleDB(Parser.DBPath)
    if Parser.LoadDistributionPath:
        print "Load p-value distribution from %s..."%Parser.LoadDistributionPath
        Parser.LoadPValueDistribution(Parser.LoadDistributionPath)
    elif Parser.ReadScoresPath:
        print "Read scores from search results at %s..."%Parser.ReadScoresPath
        Parser.ReadDeltaScoreDistribution(Parser.ReadScoresPath)
        Parser.SetOutputDistributionPath(Parser.SaveDistributionPath)
        ##############################
        # Loop for F-score methods
        Parser.ProcessResultsFiles(Parser.ReadScoresPath, Parser.ReadScoreDistributionFromFile, MAX_RESULTS_FILES_TO_PARSE)
        if Parser.ShuffledDatabaseFraction != None:
            print "Compute PValues with shuffled..."
            Parser.ComputePValuesWithShuffled(0)
            Parser.ComputePValuesWithShuffled(1)
        else:
            Result = Parser.FitMixtureModel()
            if not Result:
                sys.exit(1)
        if Parser.PerformProteinSelection:
            Parser.SelectProteins(Parser.PValueCutoff, Parser.ReadScoresPath)
        print "Write p-value distribution to %s..."%Parser.SaveDistributionPath
        (Stub, Extension) = os.path.splitext(Parser.SaveDistributionPath)
        Parser.SavePValueDistribution(0)
        Parser.SavePValueDistribution(1)
        ##############################
        Parser.OutputDistributionFile.close()
    else:
        print "** Please specify either a distribution file or results file."
        print UsageInfo
        sys.exit(1)
    if Parser.GenerateImageFlag and Image:
        ImagePath = os.path.splitext(Parser.SaveDistributionPath)[0] + ".2.png"
        Parser.ProduceScoreDistributionImage(ImagePath, 0)
        ImagePath = os.path.splitext(Parser.SaveDistributionPath)[0] + ".3.png"
        Parser.ProduceScoreDistributionImage(ImagePath, 1)
    if Parser.WriteScoresPath:
        Parser.WriteFixedScores(Parser.WriteScoresPath)
if __name__ == "__main__":
    try:
        import psyco
        psyco.full()
    except:
        print "psyco not found - running without optimization"
    #TestMain()
    Main()



#Title:          CombinePTMFeatures.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
CombinePTMFeatures:
- Parse the output of several runs of ComputePTMFeatures.
- Accumulate consensus clusters, and accumulate coverage
- Write the results to a merged directory
"""
import sys
import os
import math
import getopt
import string
import struct
import traceback
import shutil
import BuildConsensusSpectrum
import ResultsParser
import PyInspect
from Utils import *
from TrainPTMFeatures import FormatBits

UsageInfo = """
-r [PATH]: Directory containing the results-directories to merge
-w [PATH]: Output file
-d [PATH]: Database path

Optional:
-M [DIR]: Directory subtree where mzxml files *really* live.
    Spectrum paths will be corrected to use these paths.
-x: If set, prepare output directories.  (Should be set for the first
    run of a batch, and NOT for any others)
-s [POS]: Start DBPosition
-e [POS]: End DBPosition
-q: Quick-parse flag
-c [STRING]: Required filename chunk
"""

class PTMFeatureMerger(ResultsParser.SpectrumOracleMixin):
    def __init__(self):
        self.OutputDir = None
        self.Peptides = {} # (annotation, charge) -> peptide species
        self.HeaderLines = []
        self.HeaderLinesParsed = 0
        self.DBStart = None
        self.DBEnd = None
        self.QuickParseFlag = 0
        self.TotalSpectrumCount = 0
        self.SpectrumRoot = None
        ResultsParser.SpectrumOracleMixin.__init__(self)
    def WipeDir(self, Dir):
        try:
            shutil.rmtree(Dir)
        except:
            pass
    def ParseCommandLine(self, Arguments):
        PrepareDirsFlag = 0
        (Options, Args) = getopt.getopt(Arguments, "d:r:w:s:e:qxM:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-r":
                self.PTMFeatureDirectory = Value
            elif Option == "-d":
                self.DBPath = Value
            elif Option == "-w":
                self.OutputPath = Value
            elif Option == "-s":
                self.DBStart = int(Value)
            elif Option == "-e":
                self.DBEnd = int(Value)
            elif Option == "-q":
                self.QuickParseFlag = 1
            elif Option == "-x":
                PrepareDirsFlag = 1
            elif Option == "-M":
                self.SpectrumRoot = Value
            else:
                print "* Error: Unrecognized option %s"%Option
        if not self.OutputPath:
            print "* Please specify an output file (-w)"
            print UsageInfo
            sys.exit(-1)
        self.OutputDir = os.path.split(self.OutputPath)[0]
        #self.OutputPath = os.path.join(self.OutputDir, "PTMFeatures.txt")
        self.ClusterDir = os.path.join(self.OutputDir, "Clusters")
        self.SpectrumDir = os.path.join(self.OutputDir, "Spectra")
        self.ClusterMemberDir = os.path.join(self.OutputDir, "ClusterMembers")
        print "Prepare directories..."
        if PrepareDirsFlag:
            self.WipeDir(self.OutputDir)
            MakeDirectory(self.OutputDir)
            for Dir in (self.ClusterDir, self.SpectrumDir, self.ClusterMemberDir):
                MakeDirectory(Dir)
                for AA in "ACDEFGHIKLMNPQRSTVWY":
                    MakeDirectory(os.path.join(Dir, AA))
        return 1 # success
    def LoadDB(self):
        # Populate self.DB with the contents of the .trie file
        File = open(self.DBPath, "rb")
        self.DB = File.read()
        File.close()
        self.Coverage = [0] * len(self.DB)
        self.ModCoverage = [0] * len(self.DB)
    def OutputCoverage(self):
        CoveragePath = os.path.join(self.OutputDir, "Coverage.dat")
        CoverageFile = open(CoveragePath, "wb")
        for DBPos in range(len(self.DB)):
            Str = struct.pack("<II", self.Coverage[DBPos], self.ModCoverage[DBPos])
            CoverageFile.write(Str)
        CoverageFile.close()
    def OutputPTMFeatures(self):
        File = open(self.OutputPath, "wb")
        for FileLine in self.HeaderLines:
            File.write(FileLine)
        for Peptide in self.Peptides.values():
            String = string.join(Peptide.Bits, "\t")
            File.write(String + "\n")
    def GrabClusterMembers(self, Directory, Peptide):
        InputPath = os.path.join(Directory, "ClusterMembers", Peptide.Annotation[2], "%s.%s.txt"%(Peptide.Annotation, Peptide.Charge))
        OutputPath = os.path.join(self.ClusterMemberDir, Peptide.Annotation[2], "%s.%s.txt"%(Peptide.Annotation, Peptide.Charge))
        InputFile = open(InputPath, "rb")
        OutputFile = open(OutputPath, "a+b")
        OutputFile.write(InputFile.read())
        InputFile.close()
        OutputFile.close()
    def BuildNewPeptide(self, Cursor):
        """
        We've parsed the first of the cursors that contains this peptide species.  Build a peptide
        and start accumulating clusters.
        """
        Bits = Cursor.Bits
        
    def AddNewPeptide(self, Directory, Bits):
        """
        We're parsing a peptide (from the ComputePTMFeatures output) which we haven't
        seen before.  Create a Peptide object and populate it.
        """
        Peptide = Bag()
        Peptide.Bits = list(Bits)
        Peptide.Charge = int(Bits[FormatBits.Charge])
        Peptide.Annotation = Bits[FormatBits.Peptide]
        SpectrumCount = int(Bits[FormatBits.SpectrumCount])
        self.TotalSpectrumCount += SpectrumCount
        Peptide.Peptide = GetPeptideFromModdedName(Peptide.Annotation)
        Peptide.ModlessAnnotation = "%s.%s.%s"%(Peptide.Peptide.Prefix, Peptide.Peptide.Aminos, Peptide.Peptide.Suffix)
        # Grab the cluster members right away:
        self.GrabClusterMembers(Directory, Peptide)
        # Add a cluster for the peptide:
        Peptide.Cluster = BuildConsensusSpectrum.ConsensusBuilder(Peptide.Charge)
        ClusterPath = os.path.join(Directory, "Clusters", Peptide.Annotation[2], "%s.%s.cls"%(Peptide.Annotation, Peptide.Charge))
        Peptide.Cluster.UnpickleCluster(ClusterPath)
        # Add a cluster for the modless peptide, if the cluster-file exists:
        ModlessClusterPath = os.path.join(Directory, "Clusters", Peptide.Annotation[2], "%s.%s.cls"%(Peptide.ModlessAnnotation, Peptide.Charge))
        if os.path.exists(ModlessClusterPath):
            Peptide.ModlessCluster = BuildConsensusSpectrum.ConsensusBuilder(Peptide.Charge)
            Peptide.ModlessCluster.UnpickleCluster(ModlessClusterPath)
        else:
            Peptide.ModlessCluster = None
        # Add the peptide to our dictionary:
        Key = (Peptide.Annotation, Peptide.Charge)
        self.Peptides[Key] = Peptide
    def AssimilatePeptide(self, Dir, Peptide, Bits):
        """
        We're parsing a peptide (from the ComputePTMFeatures output) which we 
        have already seen.  Adjust the features of our Peptide object - accumulate
        spectrum counts, et cetera.
        """
        # Best modless spectrum and MQScore.  These may be empty for the new
        # file bits, for the existing peptide, or both.
        ScoreStr = Peptide.Bits[FormatBits.BestModlessMQScore]
        if ScoreStr:
            OldModlessMQScore = float(ScoreStr)
        else:
            OldModlessMQScore = None
        ScoreStr = Bits[FormatBits.BestModlessMQScore]
        if ScoreStr:
            ModlessMQScore = float(ScoreStr)
        else:
            ModlessMQScore = None
        if ModlessMQScore > OldModlessMQScore:
            Peptide.Bits[FormatBits.BestModlessMQScore] = Bits[FormatBits.BestModlessMQScore]
            Peptide.Bits[FormatBits.BestModlessSpectrumPath] = Bits[FormatBits.BestModlessSpectrumPath]
        # Best modded spectrum, mqscore, delta-score:
        OldBestMQScore = float(Peptide.Bits[FormatBits.BestMQScore])
        BestMQScore = float(Bits[FormatBits.BestMQScore])
        if BestMQScore > OldBestMQScore:
            Peptide.Bits[FormatBits.BestMQScore] = Bits[FormatBits.BestMQScore]
            Peptide.Bits[FormatBits.BestDeltaScore] = Bits[FormatBits.BestDeltaScore]
            Peptide.Bits[FormatBits.BestSpectrumPath] = Bits[FormatBits.BestSpectrumPath]
        # Spectra:
        CurrentSpectra = int(Peptide.Bits[FormatBits.SpectrumCount])
        NewBlockSpectra = int(Bits[FormatBits.SpectrumCount])
        TotalSpectra = CurrentSpectra + NewBlockSpectra
        self.TotalSpectrumCount += NewBlockSpectra
        Peptide.Bits[FormatBits.SpectrumCount] = str(Spectra)
        # Modless spectra
        Spectra = int(Peptide.Bits[FormatBits.ModlessSpectrumCount])
        Spectra += int(Bits[FormatBits.ModlessSpectrumCount])
        Peptide.Bits[FormatBits.ModlessSpectrumCount] = str(Spectra)
        # Accumulate modded spectra into the cluster:
        ClusterPath = os.path.join(Dir, "Clusters", Peptide.Annotation[2], "%s.%s.cls"%(Peptide.Annotation, Peptide.Charge))
        TempCluster = BuildConsensusSpectrum.ConsensusBuilder(Peptide.Charge)
        TempCluster.UnpickleCluster(ClusterPath)
        Peptide.Cluster.AssimilateCluster(TempCluster)
        # Accumulate modless spectra into the modless cluster:
        ClusterPath = os.path.join(Dir, "Clusters", Peptide.Annotation[2], "%s.%s.cls"%(Peptide.ModlessAnnotation, Peptide.Charge))
        if os.path.exists(ClusterPath):
            TempCluster = BuildConsensusSpectrum.ConsensusBuilder(Peptide.Charge)
            TempCluster.UnpickleCluster(ClusterPath)
            if Peptide.ModlessCluster:
                Peptide.ModlessCluster.AssimilateCluster(TempCluster)
            else:
                Peptide.ModlessCluster = TempCluster
        # Consensus MQScore handled at the end
        # Spectra/sites this mod type handled at the *VERY* end, possibly after
        #   multiple runs of CombinePTMFeatures!
        # Log spectrum-count handled at the end
    def MergeResultsFromFile(self, Path):
        Dir = os.path.split(Path)[0]
        File = open(Path, "rb")
        LineNumber = 0
        for FileLine in File.xreadlines():
            LineNumber += 1
            if LineNumber % 100 == 0:
                print "%s line %s (%s peptides, %s spectra)"%(Path, LineNumber, len(self.Peptides.keys()), self.TotalSpectrumCount)
                if self.QuickParseFlag:
                    break
            if FileLine[0] == "#":
                if not self.HeaderLinesParsed:
                    self.HeaderLines.append(FileLine)
                continue
            Bits = FileLine.strip().split("\t")
            # Skip any blank lines:
            if len(Bits) < 2:
                continue
            try:
                DBPos = int(Bits[FormatBits.DBPos])
                Annotation = Bits[FormatBits.Peptide]
                Charge = int(Bits[FormatBits.Charge])
            except:
                print "* Warning: Line %s of %s isn't valid!"%(LineNumber, Path)
                traceback.print_exc()
                continue
            # Ignore any peptides which don't fall within our database region of interest:
            if self.DBStart != None and DBPos < self.DBStart:
                continue
            if self.DBEnd != None and DBPos >= self.DBEnd:
                continue
            Key = (Annotation, Charge)
            Peptide = self.Peptides.get(Key, None)
            if Peptide:
                self.AssimilatePeptide(Dir, Peptide, Bits)
            else:
                self.AddNewPeptide(Dir, Bits)
        File.close()
        self.HeaderLinesParsed = 1
    def FinalizePTMFeatures(self):
        """
        Some PTM feature processing, such as building and scoring a consensus spectrum,
        should happen just once.  Those steps happen here, *after* each input file
        has been parsed.
        """
        for Peptide in self.Peptides.values():
            # Update log-spectrum-count:
            Spectra = int(Peptide.Bits[FormatBits.SpectrumCount])
            Peptide.Bits[FormatBits.LogSpectrumCount] = str(math.log(Spectra))
            # Write out the consensus MODLESS cluster:
            if Peptide.ModlessCluster:
                Path = os.path.join(self.ClusterDir, Peptide.Annotation[2], "%s.%s.cls"%(Peptide.ModlessAnnotation, Peptide.Charge))
                Peptide.ModlessCluster.PickleCluster(Path)
            # Write out the consensus MODLESS spectrum:
            if Peptide.ModlessCluster:
                Path = os.path.join(self.SpectrumDir, Peptide.Annotation[2], "%s.%s.dta"%(Peptide.ModlessAnnotation, Peptide.Charge))
                Spectrum = Peptide.ModlessCluster.ProduceConsensusSpectrum()
                Spectrum.WritePeaks(Path)
            # Write out the CLUSTER:
            Path = os.path.join(self.ClusterDir, Peptide.Annotation[2], "%s.%s.cls"%(Peptide.Annotation, Peptide.Charge))
            Peptide.Cluster.PickleCluster(Path)
            # Write out the consensus SPECTRUM:
            ConsensusSpectrumPath = os.path.join(self.SpectrumDir, Peptide.Annotation[2], "%s.%s.dta"%(Peptide.Annotation, Peptide.Charge))
            Spectrum = Peptide.Cluster.ProduceConsensusSpectrum()
            Spectrum.WritePeaks(ConsensusSpectrumPath)
            # Compute consensus spectrum features:
            PySpectrum = PyInspect.Spectrum(ConsensusSpectrumPath, 0)
            ScoreList = PySpectrum.ScorePeptideDetailed(Peptide.Annotation)
            Peptide.Bits[FormatBits.ConsensusMQScore] = str(ScoreList[0])
            Peptide.Bits[FormatBits.ConsensusPRMScore] = str(ScoreList[1])
            Peptide.Bits[FormatBits.ConsensusBYPresent] = str(ScoreList[2])
            Peptide.Bits[FormatBits.ConsensusTopPeaks] = str(ScoreList[3])
            Peptide.Bits[FormatBits.NTT] = str(ScoreList[4])
            # Compute comparison features:
            if Peptide.ModlessCluster:
                Peptide.Bits[FormatBits.SisterAnnotationFlag] = "1"
                pass #NOTE: skip these features, since we don't really use them!
            
    def AssimilateDatabaseCoverage(self, CoverageFilePath):
        """
        Read Coverage.dat from one of the ComputePTMFeatures runs.
        Accumulate total coverage.
        """
        StructSize = struct.calcsize("<II")
        File = open(CoverageFilePath, "rb")
        for DBPos in range(len(self.DB)):
            Block = File.read(StructSize)
            (Coverage, ModCoverage) = struct.unpack("<II", Block)
            self.Coverage[DBPos] += Coverage
            self.ModCoverage[DBPos] += ModCoverage
            if DBPos % 10000 == 0:
                print "%s/%s..."%(DBPos, len(self.DB))
        File.close()
    def WriteSingletonPeptide(self, Species, Cursor):
        """
        Here's a peptide that was found in only one of the input files.  That makes our job very
        easy; all we need to do is write out the fileline, and copy over: cluster members,
        cluster file, spectrum file, modless cluster file (if it exists), modless spectrum file (if
        it exists)
        """
        # Write the file line:
        OutputLine = string.join(Cursor.Bits, "\t")
        self.OutputFile.write(OutputLine + "\n")
        # Copy the needful files:
        Annotation = Species.Annotation
        Charge = Species.Charge
        AA = Species.AA
        SourcePath = os.path.join(Cursor.Directory, "ClusterMembers", AA, "%s.%s.txt"%(Annotation, Charge))
        TargetPath = os.path.join(self.OutputDir, "ClusterMembers", AA, "%s.%s.txt"%(Annotation, Charge))
        shutil.copyfile(SourcePath, TargetPath)
        SourcePath = os.path.join(Cursor.Directory, "Clusters", AA, "%s.%s.cls"%(Annotation, Charge))
        TargetPath = os.path.join(self.OutputDir, "Clusters", AA, "%s.%s.cls"%(Annotation, Charge))
        shutil.copyfile(SourcePath, TargetPath)
        SourcePath = os.path.join(Cursor.Directory, "Spectra", AA, "%s.%s.dta"%(Annotation, Charge))
        TargetPath = os.path.join(self.OutputDir, "Spectra", AA, "%s.%s.dta"%(Annotation, Charge))
        shutil.copyfile(SourcePath, TargetPath)
        SourcePath = os.path.join(Cursor.Directory, "Clusters", AA, "%s.%s.cls"%(Species.ModlessAnnotation, Charge))
        TargetPath = os.path.join(self.OutputDir, "Clusters", AA, "%s.%s.cls"%(Species.ModlessAnnotation, Charge))
        if os.path.exists(SourcePath):
            shutil.copyfile(SourcePath, TargetPath)
        SourcePath = os.path.join(Cursor.Directory, "Spectra", AA, "%s.%s.dta"%(Species.ModlessAnnotation, Charge))
        TargetPath = os.path.join(self.OutputDir, "Spectra", AA, "%s.%s.dta"%(Species.ModlessAnnotation, Charge))
        if os.path.exists(SourcePath):
            shutil.copyfile(SourcePath, TargetPath)
    def BuildPeptideSpecies(self, Cursor):
        Species = Bag()
        Bits = Cursor.Bits
        Species.Annotation = Bits[FormatBits.Peptide]
        Species.AA = Species.Annotation[2]
        Species.Peptide = GetPeptideFromModdedName(Species.Annotation)
        Species.ModlessAnnotation = "%s.%s.%s"%(Species.Peptide.Prefix, Species.Peptide.Aminos, Species.Peptide.Suffix)
        Species.Charge = int(Bits[FormatBits.Charge])
        Species.Bits = Bits[:]
        Species.Cluster = None # ConsensusBuilder, instantaited later
        Species.ModlessCluster = None # ConsensusBuilder, instantaited later
        # Fix the spectrum PATHS:
        Species.Bits[FormatBits.BestSpectrumPath] = self.FixSpectrumPath(Species.Bits[FormatBits.BestSpectrumPath])
        if Species.Bits[FormatBits.BestModlessSpectrumPath]:
            Species.Bits[FormatBits.BestModlessSpectrumPath] = self.FixSpectrumPath(Species.Bits[FormatBits.BestModlessSpectrumPath])
        return Species
    def AssimilatePeptideSpectra(self, Species, Cursor):
        """
        Accumulate total spectra for this species.
        """
        ###############################
        # Adjust features - best spectrum, number of spectra, etc.
        # Best modless spectrum and MQScore.  These may be empty for the new
        # file bits, for the existing peptide, or both.
        ScoreStr = Species.Bits[FormatBits.BestModlessMQScore]
        if ScoreStr:
            OldModlessMQScore = float(ScoreStr)
        else:
            OldModlessMQScore = None
        ScoreStr = Cursor.Bits[FormatBits.BestModlessMQScore]
        if ScoreStr:
            ModlessMQScore = float(ScoreStr)
        else:
            ModlessMQScore = None
        if ModlessMQScore > OldModlessMQScore:
            Species.Bits[FormatBits.BestModlessMQScore] = Cursor.Bits[FormatBits.BestModlessMQScore]
            Species.Bits[FormatBits.BestModlessSpectrumPath] = Cursor.Bits[FormatBits.BestModlessSpectrumPath]
        # Best modded spectrum, mqscore, delta-score:
        OldBestMQScore = float(Species.Bits[FormatBits.BestMQScore])
        BestMQScore = float(Cursor.Bits[FormatBits.BestMQScore])
        if BestMQScore > OldBestMQScore:
            Species.Bits[FormatBits.BestMQScore] = Cursor.Bits[FormatBits.BestMQScore]
            Species.Bits[FormatBits.BestDeltaScore] = Cursor.Bits[FormatBits.BestDeltaScore]
            Species.Bits[FormatBits.BestSpectrumPath] = Cursor.Bits[FormatBits.BestSpectrumPath]
        # Spectra:
        CurrentSpectra = int(Species.Bits[FormatBits.SpectrumCount])
        NewBlockSpectra = int(Cursor.Bits[FormatBits.SpectrumCount])
        TotalSpectra = CurrentSpectra + NewBlockSpectra
        self.TotalSpectrumCount += NewBlockSpectra
        Species.Bits[FormatBits.SpectrumCount] = str(TotalSpectra)
        # Log of spectrum-count:
        Species.Bits[FormatBits.LogSpectrumCount] = str(math.log(TotalSpectra))
        # Modless spectra
        Spectra = int(Species.Bits[FormatBits.ModlessSpectrumCount])
        Spectra += int(Cursor.Bits[FormatBits.ModlessSpectrumCount])
        Species.Bits[FormatBits.ModlessSpectrumCount] = str(Spectra)
        ###############################
        # Accumulate a list of cluster members:
        ClusterMemberFileName = "%s.%s.txt"%(Species.Annotation, Species.Charge)
        ClusterMemberPath = os.path.join(self.OutputDir, "ClusterMembers", Species.AA, ClusterMemberFileName)
        ClusterMemberFile = open(ClusterMemberPath, "a+b")
        CursorMemberPath = os.path.join(Cursor.Directory, "ClusterMembers", Species.AA, ClusterMemberFileName)
        CursorMemberFile = open(CursorMemberPath, "rb") 
        Text = CursorMemberFile.read()
        ClusterMemberFile.write(Text)
        CursorMemberFile.close()
        ClusterMemberFile.close()
        ###############################
        # Accumulate members of the modded cluster:
        ClusterPath = os.path.join(Cursor.Directory, "Clusters", Species.AA, "%s.%s.cls"%(Species.Annotation, Species.Charge))
        CursorCluster = BuildConsensusSpectrum.ConsensusBuilder(Species.Charge)
        CursorCluster.UnpickleCluster(ClusterPath)
        if not Species.Cluster:
            Species.Cluster = CursorCluster 
        else:
            Species.Cluster.AssimilateCluster(CursorCluster)
        ###############################
        # Accumulate members of the modless cluster:
        ClusterPath = os.path.join(Cursor.Directory, "Clusters", Species.AA, "%s.%s.cls"%(Species.ModlessAnnotation, Species.Charge))
        if os.path.exists(ClusterPath):
            CursorCluster = BuildConsensusSpectrum.ConsensusBuilder(Species.Charge)
            CursorCluster.UnpickleCluster(ClusterPath)
            if not Species.ModlessCluster:
                Species.ModlessCluster = CursorCluster 
            else:
                Species.ModlessCluster.AssimilateCluster(CursorCluster)
    def WriteCompletedPeptide(self, Species):
        """
        We've read data for this peptide species from TWO OR MORE cursors.  Now we'll write out
        one line to the output file, and output our consensus spectrum.
        """
        # Write the file line:
        OutputLine = string.join(Species.Bits, "\t")
        self.OutputFile.write(OutputLine + "\n")
        # Cluster:
        ClusterOutputPath = os.path.join(self.OutputDir, "Clusters", Species.AA, "%s.%s.cls"%(Species.Annotation, Species.Charge))
        Species.Cluster.PickleCluster(ClusterOutputPath)
        # Consensus spectrum:
        ConsensusSpectrum = Species.Cluster.ProduceConsensusSpectrum()
        ConsensusSpectrumPath = os.path.join(self.OutputDir, "Spectra", Species.AA, "%s.%s.dta"%(Species.Annotation, Species.Charge))
        ConsensusSpectrum.WritePeaks(ConsensusSpectrumPath)
        if Species.ModlessCluster:
            # Modless cluster:
            ClusterOutputPath = os.path.join(self.OutputDir, "Clusters", Species.AA, "%s.%s.cls"%(Species.ModlessAnnotation, Species.Charge))
            Species.ModlessCluster.PickleCluster(ClusterOutputPath)
            # Modless consensus spectrum:
            ConsensusSpectrum = Species.ModlessCluster.ProduceConsensusSpectrum()
            ConsensusSpectrumPath = os.path.join(self.OutputDir, "Spectra", Species.AA, "%s.%s.dta"%(Species.ModlessAnnotation, Species.Charge))
            ConsensusSpectrum.WritePeaks(ConsensusSpectrumPath)
    def MergeResults(self):
        print "Load db..."
        self.LoadDB()
        # Measure combined db coverage:
        print "Combine database coverage..."
        self.CombineDatabaseCoverage()
        print "Populate MZXML oracle..."
        self.PopulateMZXMLOracle(self.SpectrumRoot)
        self.OutputFile = open(self.OutputPath, "wb")
        for FileLine in self.HeaderLines:
            self.OutputFile.write(FileLine)        
        class FeatureCursor:
            """
            Wrapper for a feature file - tracks the "next" peptide.  
            """
            def __init__(self, Path):
                self.Path = Path
                self.File = open(Path, "rb")
                self.NextKey = None
                self.Bits = None
                self.Directory = os.path.split(Path)[0]
                self.HeaderLines = []
            def Close(self):
                self.File.close()
                self.NextKey = None
            def GetNextPeptide(self):
                # Read one or more lines (skipping header or invalid lines), and remember
                # the next peptide to be processed
                while (1):
                    FileLine = self.File.readline()
                    if not FileLine:
                        self.NextKey = None
                        return None # EOF
                    # Skip blank or comment lines:
                    if FileLine[0] == "#":
                        self.HeaderLines.append(FileLine)
                        continue
                    if not FileLine.strip():
                        continue
                    # Attempt to parse the line:
                    Bits = FileLine.strip().split("\t")
                    try:
                        ModDBPos = int(Bits[FormatBits.DBPos])
                        ModMass = int(Bits[FormatBits.ModificationMass])
                        Annotation = Bits[FormatBits.Peptide]
                        Charge = int(Bits[FormatBits.Charge])
                    except:
                        traceback.print_exc()
                        continue # skip invalid line
                    # We know our next key, so stop now:
                    self.NextKey = (ModDBPos, ModMass, Annotation, Charge)
                    self.Bits = Bits
                    break
        self.FeatureCursors = []
        # List the directories that need to be parsed, and open the files:
        for SubDirectory in os.listdir(self.PTMFeatureDirectory):
            Dir = os.path.join(self.PTMFeatureDirectory, SubDirectory)
            if not os.path.isdir(Dir):
                continue
            FeatureFilePath = os.path.join(Dir, "PTMFeatures.txt")
            if not os.path.exists(FeatureFilePath):
                print "* Warning: Subdirectory %s doesn't contain a feature file!"%Dir
                continue
            CoverageFilePath = os.path.join(Dir, "Coverage.dat")
            if not os.path.exists(CoverageFilePath):
                print "* Warning: Subdirectory %d doesn't contain a coverage file!"%Dir
                continue
            Cursor = FeatureCursor(FeatureFilePath)
            Cursor.GetNextPeptide()
            self.FeatureCursors.append(Cursor)
        # Output the header lines from an (arbitrary) cursor:
        for HeaderLine in self.FeatureCursors[0].HeaderLines:
            self.OutputFile.write(HeaderLine)
        # Loop through the peptides, until all cursors hit EOF.  At each stage, process
        # the peptide with the first key - and process it from any and all cursors which
        # are now pointing at it.
        while (1):
            CursorsForThisKey = 0
            CursorThisKey = None
            FirstKey = None
            print 
            for Cursor in self.FeatureCursors:
                if Cursor.NextKey != None and (FirstKey == None or FirstKey > Cursor.NextKey):
                    FirstKey = Cursor.NextKey
                    CursorThisKey = Cursor
                    CursorsForThisKey = 0
                if Cursor.NextKey == FirstKey:
                    CursorsForThisKey += 1
                if Cursor.NextKey == FirstKey:
                    print "*Cursor %s: %s"%(Cursor.Path, Cursor.NextKey) 
                else:
                    print "Cursor %s: %s"%(Cursor.Path, Cursor.NextKey) 
            if FirstKey == None:
                break
            Species = self.BuildPeptideSpecies(CursorThisKey)
            print "Next species is '%s', found in %s files."%(Species.Annotation, CursorsForThisKey)
            ########################################################################
            # Shortcut: If CursorsForThisKey == 1, then we don't need to re-build the
            # cluster or the consensus spectrum!
            if CursorsForThisKey == 1:
                self.WriteSingletonPeptide(Species, CursorThisKey)
                CursorThisKey.GetNextPeptide()
                continue
            ########################################################################
            # Standard case: Two or more cursors have the same peptide.  Assimilate the
            # total spectra, and cluster members, from each one!
            for Cursor in self.FeatureCursors:
                if Cursor.NextKey == FirstKey:
                    self.AssimilatePeptideSpectra(Species, Cursor)
                    Cursor.GetNextPeptide()
            # We've assimilated all spectra into the cluster; now write out the
            # totaled spectrum count and the consensus spectra!
            self.WriteCompletedPeptide(Species)
        ########################################################
        # All peptides have been written out; every cursor is at EOF.
        # Loop over cursors and finish up:
        for Cursor in self.FeatureCursors:
            Cursor.Close()
    def CombineDatabaseCoverage(self):
        """
        Iterate over directories, and compute the total coverage for each database residue.
        """
        for SubDirectory in os.listdir(self.PTMFeatureDirectory):
            Dir = os.path.join(self.PTMFeatureDirectory, SubDirectory)
            if not os.path.isdir(Dir):
                continue
            CoverageFilePath = os.path.join(Dir, "Coverage.dat")
            if not os.path.exists(CoverageFilePath):
                print "* Warning: Subdirectory %s doesn't contain a coverage file!"%Dir
                continue
            print "Assimilate from %s..."%CoverageFilePath
            self.AssimilateDatabaseCoverage(CoverageFilePath)
        self.OutputCoverage()

if __name__ == "__main__":
    try:
        import psyco
    except:
        print "(psyco not available - no optimization for you)"
    Merger = PTMFeatureMerger()
    Result = Merger.ParseCommandLine(sys.argv[1:])
    if not Result:
        print UsageInfo
        sys.exit(-1)
    Merger.MergeResults()
    



#Title:          ComputePTMFeatures.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
Plan:
Output a large collection of features for each post-translational modification accepted on a
search of a part-bogus database.  All modifications on the bogus proteins are incorrect.
An equivalent number of modifications on the non-bogus proteins are incorrect.  Let's compute
a variety of features for the PTMs observed.

Input:
A collection of annotated spectra, output by SelectSites.py
Output:
A file listing all the observed modification sites, with various features computed.  

Then, we train a model to distinguish between good (correct DB) and bad (incorrect DB)
modifications.  Model types: LDA, logistic regression, SVM, etc.

(Another possible experiment: Search unmodified spectra against a mutated database,
judge correct precisely those modifications which "undo" the mutations)
"""

import os
import sys
import struct
import traceback
import getopt
import MSSpectrum
import PyInspect
import random
import shutil
import time
import math
import cPickle
import BasicStats
import ResultsParser
import BuildConsensusSpectrum
import SpectralSimilarity
import StripPTM
random.seed(1)
from Utils import *
from TrainPTMFeatures import FormatBits
Initialize()

AMINO_ACIDS = "ACDEFGHIKLMNOPQRSTUVWY" # O and U are included, for now.
INVALID_MASS = 99999

# Retain at most this many spectra for an unmodified peptide:
MAX_MODLESS_CLUSTER_SIZE = 100

# For running the Python profiler:
PROFILE_FLAG = 0

class PeptideSpecies:
    """
    Peptides[(Annotation, Charge)] -> PeptideSpecies instance
    The PeptideSpecies remembers a list of spectra, the modification position, the modification mass.
    """
    InstanceCount = 0
    def __init__(self):
        self.HitCount = 0
        self.ModifiedFlag = 0
        self.ModMass = 0
        self.DBPos = 0
        self.ModDBPos = 0
        self.Spectra = []
        self.Peptide = None
        self.SpectrumCount = 0
        PeptideSpecies.InstanceCount += 1
    def __del__(self):
        if PeptideSpecies:
            PeptideSpecies.InstanceCount -= 1
    def __str__(self):
        return self.Annotation

class SpectrumInfoClass(ResultsParser.ResultsParser):
    """
    Information about a single scan.  We remember only the info we'll need later.
    """
    InstanceCount = 0
    def __init__(self, Bits, Trainer):
        ResultsParser.ResultsParser.__init__(self)
        
        self.FileNameIndex = Trainer.RememberString(Trainer.CachedFilePaths, Bits[0])
        self.MQScore = float(Bits[Trainer.Columns.getIndex("MQScore")])
        self.DeltaScore = float(Bits[Trainer.Columns.getIndex("DeltaScore")])
        self.ByteOffset = int(Bits[Trainer.Columns.getIndex("SpecFilePos")])
        self.ScanNumber = int(Bits[Trainer.Columns.getIndex("Scan#")])
        SpectrumInfoClass.InstanceCount += 1
    def __cmp__(self, Other):
        """
        Sort from BEST to WORST match.
        """
        if self.MQScore > Other.MQScore:
            return -1
        if self.MQScore < Other.MQScore:
            return 1
        return 0
    def __del__(self):
        if SpectrumInfoClass:
            SpectrumInfoClass.InstanceCount -= 1
        
class PTMFeatureComputer(ResultsParser.ResultsParser, ResultsParser.SpectrumOracleMixin):
    def __init__(self):
        self.PValueCutoff = 0.1 # default
        self.ResultsFileName = None
        self.DBPath = None
        self.OutputDir = "PTM"
        self.OutputPath = os.path.join(self.OutputDir, "PTMFeatures.txt")
        self.ConsensusClusterDir = None
        self.ConsensusSpectrumDir = None
        # Peptides keeps a list of SpectrumInfo objects for each peptide species
        # we've observed.  The keys have the form (Annotation, Charge) and the values
        # are lists SpectrumInfo instance.
        self.Peptides = {}
        self.PTMs = {} # keys of the form (DBPos, Mass)
        self.CoverageThreshold = 2 # at least this many spectra to consider a residue 'covered'.
        self.QuickParseFlag = 0 # if true, then parse only the first n lines
        self.PoolFlag = 0
        self.ModelType = None
        self.SisterProteins = {} # protein index -> sister protein's index
        self.MZXMLOracle = {}
        self.ModelTrainFilePath = "PTMFeatures.All.txt"
        self.ModelTestFilePath = None
        # Dictionary of unmodified peptides, for computing the coverage level:
        self.UnmodifiedPeptides = {}
        self.FeatureSelectionFlag = None
        self.CachedFilePaths = []
        self.CachedFixedFilePaths = []
        self.StartOutputDBPos = 0
        self.RequiredFileNameChunk = None
        self.Columns = ResultsParser.Columns()
        ResultsParser.ResultsParser.__init__(self)
        ResultsParser.SpectrumOracleMixin.__init__(self)

    def RememberString(self, StringList, NewString):
        """
        Return the index of NewString within StringList, adding to the list if necessary.
        We keep a list of mzxml file names and store indexes into the list, to avoid
        the memory hit required to store each occurrence of the name.
        """
        try:
            Index = StringList.index(NewString)
            return Index
        except:
            StringList.append(NewString)
            return len(StringList) - 1
    def LoadDB(self):
        """
        Load the database searched.  For future reference, we want the protein names as well.
        """
        # Populate self.DB with the contents of the .trie file
        File = open(self.DBPath, "rb")
        self.DB = File.read()
        File.close()
        # Populate self.ProteinNames and self.ProteinPositions by parsing the index file:
        self.ProteinNames = []
        self.ProteinPositions = []
        IndexPath = os.path.splitext(self.DBPath)[0] + ".index"
        File = open(IndexPath, "rb")
        BlockSize = struct.calcsize("<qi80s")
        while 1:
            Block = File.read(BlockSize)
            if not Block:
                break
            Tuple = struct.unpack("<qi80s", Block)
            Name = Tuple[-1]
            NullPos = Name.find("\0")
            if NullPos != -1:
                Name = Name[:NullPos]
            self.ProteinNames.append(Name)
            self.ProteinPositions.append(Tuple[1])
        File.close()
        # Initialize our coverage arrays:
        self.Coverage = [0] * len(self.DB)
        self.ModCoverage = [0] * len(self.DB)
        self.PeptideCoverage = [0] * len(self.DB)
        self.ModPeptideCoverage = [0] * len(self.DB)
        # Find sister-proteins.
        # A shuffled protein name should be the standard protein's name
        # with the characters "XXX" prepended.  (If the standard protein name
        # is very long, the last characters may "slide off the edge")
        for IndexA in range(len(self.ProteinNames)):
            Name = self.ProteinNames[IndexA]
            SisterName = None
            if Name[:3] == "XXX":
                SisterName = Name[3:]
            if SisterName:
                for IndexB in range(len(self.ProteinNames)):
                    Name = self.ProteinNames[IndexB][:77]
                    if Name == SisterName:
                        self.SisterProteins[IndexA] = IndexB
                        self.SisterProteins[IndexB] = IndexA
    def ComputeProteinCoverage(self):
        """
        Compute residue-level coverage.  And, compute
        what fraction of each protein is covered. 
        """
        for Species in self.Peptides.values():
            DBPos = Species.DBPos
            if Species.ModifiedFlag:
                for Pos in range(DBPos, DBPos + len(Species.Peptide.Aminos)):
                    self.ModCoverage[Pos] += Species.SpectrumCount # count modified spectra
                    self.ModPeptideCoverage[Pos] += 1 # count distinct peptide species
            else:
                for Pos in range(DBPos, DBPos + len(Species.Peptide.Aminos)):
                    self.Coverage[Pos] += Species.SpectrumCount # count unmodified spectra
                    self.PeptideCoverage[Pos] += 1 # count distinct unmodified peptide species
        #########################################################
        # Compute percentage of each protein that's covered:
        self.ProteinCoverageLevels = []
        for ProteinIndex in range(len(self.ProteinPositions)):
            StartPos = self.ProteinPositions[ProteinIndex]
            if ProteinIndex < len(self.ProteinPositions) - 1:
                EndPos = self.ProteinPositions[ProteinIndex + 1]
            else:
                EndPos = len(self.DB)
            #print "Protein %s (%s) from %s-%s"%(ProteinIndex, self.ProteinNames[ProteinIndex], StartPos, EndPos)
            CoverFlags = 0
            for Pos in range(StartPos, EndPos):
                if self.Coverage[Pos] >= self.CoverageThreshold:
                    CoverFlags += 1
            ProteinLength = EndPos - StartPos
            #print "  -> Coverage %s/%s = %s"%(CoverFlags, ProteinLength, CoverFlags / float(ProteinLength))
            self.ProteinCoverageLevels.append(CoverFlags / float(ProteinLength))
        # SAVE protein coverage levels:
        CoveragePath = os.path.join(self.OutputDir, "Coverage.dat")
        CoverageFile = open(CoveragePath, "wb")
        for DBPos in range(len(self.DB)):
            Str = struct.pack("<II", self.Coverage[DBPos], self.ModCoverage[DBPos])
            CoverageFile.write(Str)
        CoverageFile.close()
        # Boost protein coverage levels based upon sister proteins:
        for Index in range(len(self.ProteinNames)):
            SisterIndex = self.SisterProteins.get(Index, None)
            if SisterIndex == None:
                continue
            self.ProteinCoverageLevels[Index] = max(self.ProteinCoverageLevels[Index], self.ProteinCoverageLevels[SisterIndex])
            #print "%s and %s are sisters, with coverage %s"%(self.ProteinNames[Index], self.ProteinNames[SisterIndex], self.ProteinCoverageLevels[Index])
    def FixPeptideSpecies(self):
        """
        Iterate over all the peptide species we observed.  Strip "obviously unnecessary" PTMs.
        """
        Keys = self.Peptides.keys()
        for Key in Keys:
            Annotation = Key[0]
            Species = self.Peptides[Key]
            Result = StripPTM.StripNeedlessModifications(self.DB, Annotation)
            if not Result:
                continue
            (DBPos, FixedAnnotation) = Result
            # If the annotation wasn't changed, then continue.
            if FixedAnnotation == Annotation:
                continue
            Species.Peptide = GetPeptideFromModdedName(FixedAnnotation)
            Species.Annotation = FixedAnnotation
            Species.DBPos = DBPos
            ModKeys = Species.Peptide.Modifications.keys()
            if len(ModKeys):
                ModIndex = ModKeys[0]
                Species.ModifiedFlag = 1
                Species.ModMass = int(round(Species.Peptide.Modifications[ModIndex][0].Mass))
                Species.ModDBPos = DBPos + ModIndex
            else:
                Species.ModifiedFlag = 0
                Species.ModMass = 0
                Species.ModDBPos = None
            del self.Peptides[Key]
            # Either merge into the existing species with this fixed annotation,
            # or move into the empty pigeonhole:
            FixedKey = (FixedAnnotation, Key[1])
            OldSpecies = self.Peptides.get(FixedKey, None)
            if OldSpecies:
                OldSpecies.Spectra.extend(Species.Spectra)
                OldSpecies.SpectrumCount += Species.SpectrumCount
                if len(OldSpecies.Spectra) > MAX_MODLESS_CLUSTER_SIZE:
                    OldSpecies.Spectra.sort()
                    OldSpecies.Spectra = Species.Spectra[:MAX_MODLESS_CLUSTER_SIZE]
            else:
                self.Peptides[FixedKey] = Species
                if len(Species.Spectra) > MAX_MODLESS_CLUSTER_SIZE:
                    Species.Spectra.sort()
                    Species.Spectra = Species.Spectra[:MAX_MODLESS_CLUSTER_SIZE]
    def ParsePTMsFromResultsFile(self, FilePath):
        """
        Callback for parsing one Inspect results-file.  Our job here
        is to populate self.BestSpectra, and self.PTMs.
        Note: It's POSSIBLE that we'll spot some modified annotations
        which can be "trivially fixed" to produce unmodified annotations.
        Examples: T.T+101LAPTTVPITSAK.A, Y.E+163NPNFTGK.K
        Because of this, we keep a dictionary self.FixedAnnotation,
        where keys are raw annotations and values are fixed-up annotations.
        """
        if self.RequiredFileNameChunk:
            Pos = FilePath.find(self.RequiredFileNameChunk)
            if Pos == -1:
                return
        if os.path.isdir(FilePath):
            print "NOTE: Skipping results sub-directory '%s'"%FilePath
            return
        try:
            File = open(FilePath, "rb")
        except:
            print "** Unable to open results file '%s'"%FilePath
            return
        LineNumber = 0
        OldSpectrum = None
        for FileLine in File.xreadlines():
            LineNumber += 1
            if LineNumber % 1000 == 0:
                print "  Line %s..."%LineNumber
                if self.QuickParseFlag:
                    break
            if FileLine[0] == "#":
                self.Columns.initializeHeaders(FileLine)
                continue
            Bits = FileLine.strip().split("\t")
            if len(Bits) < 15:
                continue # not valid!
            Spectrum = (Bits[0], Bits[1])
            if Spectrum == OldSpectrum:
                continue
            OldSpectrum = Spectrum
            PValue = float(Bits[self.Columns.getIndex("InspectFDR")])
            if PValue > self.PValueCutoff:
                continue
            Annotation = Bits[self.Columns.getIndex("Annotation")]
            Charge = int(Bits[self.Columns.getIndex("Charge")])
            AnnotationKey = (Annotation, Charge)
            ##############################################################
            # If we've never seen this annotation before, then create a PeptideSpecies object
            # and record it in self.Peptides
            Species = self.Peptides.get(AnnotationKey, None)
            if not Species:
                Species = PeptideSpecies()
                Species.Peptide = GetPeptideFromModdedName(Annotation)
                Mods = []
                for (Index, List) in Species.Peptide.Modifications.items():
                    for Mod in List:
                        Mods.append((Index, Mod))
                if len(Mods):
                    Species.ModifiedFlag = 1
                    Species.ModMass = int(Mods[0][1].Mass)
                    Species.ModAA = Species.Peptide.Aminos[Mods[0][0]]
                else:
                    Species.ModifiedFlag = 0
                Species.ProteinName = Bits[self.Columns.getIndex("Protein")]
                self.Peptides[AnnotationKey] = Species
                # Get the database position of the peptide:
                Species.DBPos = self.DB.find(Species.Peptide.Aminos)
                if len(Species.Peptide.Modifications.keys()):
                    ModIndex = Species.Peptide.Modifications.keys()[0]
                    Species.ModDBPos = Species.DBPos + ModIndex
                else:
                    Species.ModDBPos = None
                # Get the residue-number of the peptide:
                StarPos = self.DB.rfind("*", 0, Species.DBPos)
                if StarPos == -1:
                    Species.ResidueNumber = Species.DBPos
                else:
                    Species.ResidueNumber = Species.DBPos - StarPos
                Species.Annotation = Annotation
                Species.Charge = Charge
            if Species.DBPos == -1:
                print "* skipping unknown peptide: %s"%Annotation
                del self.Peptides[AnnotationKey] # remove the Species that was just created!
                continue
            MQScore = float(Bits[self.Columns.getIndex("MQScore")])
            self.AnnotationCount += 1
            ##############################################################
            # Populate Species.Spectra:
            try:
                Info = SpectrumInfoClass(Bits, self)
            except:
                print "** Error: Couldn't parse spectrum info from line %s of file %s"%(LineNumber, FilePath)
                traceback.print_exc()
                continue
            Species.Spectra.append(Info)
            Species.SpectrumCount += 1
            if not Species.ModifiedFlag:
                if len(Species.Spectra) > MAX_MODLESS_CLUSTER_SIZE:
                    Species.Spectra.sort()
                    Species.Spectra = Species.Spectra[:MAX_MODLESS_CLUSTER_SIZE]
            else:
                pass
        File.close()
    def WipeDir(self, Dir):
        try:
            shutil.rmtree(Dir)
        except:
            pass
    def ComputeFeaturesMain(self):
        """
        Main method:
        - Load the database searched
        - Iterate over the results-file, to get a list of PTMs
        - Iterate over the PTMs, and write out features for each one.
        """
        self.ConsensusClusterDir = os.path.join(self.OutputDir, "Clusters")
        self.ClusterScanListDir = os.path.join(self.OutputDir, "ClusterMembers")
        self.ConsensusSpectrumDir = os.path.join(self.OutputDir, "Spectra")
        if not self.StartOutputDBPos:
            # Make sure necessary directories exist, and clean up any OLD output:
            print "Prepare cluster directories..."
            self.WipeDir(self.ConsensusClusterDir)
            self.WipeDir(self.ConsensusSpectrumDir)
            self.WipeDir(self.ClusterScanListDir)
            MakeDirectory(self.ConsensusClusterDir)
            MakeDirectory(self.ConsensusSpectrumDir)
            MakeDirectory(self.ClusterScanListDir)
            for AA in AMINO_ACIDS:
                PathA = os.path.join(self.ConsensusClusterDir, AA)
                PathB = os.path.join(self.ConsensusSpectrumDir, AA)
                PathC = os.path.join(self.ClusterScanListDir, AA)
                for Path in (PathA, PathB, PathC):
                    MakeDirectory(Path)
        else:
            print "CONTINUING ComputePTMFeatures from DBPosition %s"%self.StartOutputDBPos
        print "Load database..."
        self.LoadDB()
        print "Parse annotations..."
        self.AnnotationCount = 0
        self.PTMAnnotationCount = 0
        self.BestModlessHits = {}
        self.ProcessResultsFiles(self.ResultsFileName, self.ParsePTMsFromResultsFile)
        # Fix annotations:
        self.FixPeptideSpecies()
        # Fix file paths:
        for FilePath in self.CachedFilePaths:
            FixedPath = self.FixSpectrumPath(FilePath)
            self.CachedFixedFilePaths.append(FixedPath)
        self.PairModifiedUnmodifiedPeptides()
        print "Produce CONSENSUS SPECTRA for modified and unmodified petpides..."
        StartTime = time.clock()
        self.ProduceConsensusSpectra()
        EndTime = time.clock()
        print "Elapsed time: %s"%(EndTime - StartTime)
        print "Compute protein coverage..."
        self.ComputeProteinCoverage()
        print "Count spectra (and sites) by PTM type..."
        self.ComputeTotalSpectraForModType()
        print "Generate non-redundant PTM list..."
        self.ListDistinctPTMs()
        print "Compute features and output PTM info..."
        self.ComputeFeaturesAllPTMs()
    def PairModifiedUnmodifiedPeptides(self):
        for Species in self.Peptides.values():
            if not Species.ModifiedFlag:
                continue
            ModlessAnnotation = "%s.%s.%s"%(Species.Peptide.Prefix, Species.Peptide.Aminos, Species.Peptide.Suffix)
            ModlessKey = (ModlessAnnotation, Species.Charge)
            ModlessSpecies = self.Peptides.get(ModlessKey, None)
            Species.Modless = ModlessSpecies
    def ListDistinctPTMs(self):
        """
        Populate self.PTMs; keys are (DBPos, Mass) and values are simple objects
        with lists of peptide species.
        """
        for Species in self.Peptides.values():
            if not Species.ModifiedFlag:
                continue
            Index = Species.Peptide.Modifications.keys()[0]
            ModifiedPos = Species.DBPos + Index
            Key = (ModifiedPos, Species.ModMass)
            if not self.PTMs.has_key(Key):
                PTM = Bag()
                PTM.SpeciesList = []
                self.PTMs[Key] = PTM
                ModIndex = Species.Peptide.Modifications.keys()[0]
                PTM.DBPos = Species.ModDBPos
            PTM.SpeciesList.append(Species)
            Species.PTM = PTM
    def ComputeTotalSpectraForModType(self):
        """
        Populate a dictionary of the form (AA, Mass) -> SpectrumCount.  If a modificaiton is seen
        at multiple sites, it is more likely to be valid.
        """
        self.ModTypeSpectrumCount = {}
        self.ModTypeSiteCount = {}
        for Species in self.Peptides.values():
            if Species.ModifiedFlag:
                ModTypeKey = (Species.ModAA, Species.ModMass)
                self.ModTypeSpectrumCount[ModTypeKey] = self.ModTypeSpectrumCount.get(ModTypeKey, 0) + Species.SpectrumCount
                self.ModTypeSiteCount[ModTypeKey] = self.ModTypeSiteCount.get(ModTypeKey, 0) + 1
        pass
    def OutputPTMInfoHeader(self):
        """
        Output column headers, plus some general-purpose information such as the number
        of spectra parsed and the database size.
        """
        Header = "#Group\tDBPosition\tMass\tAminoAcid\tProtein\tResidueNumber\t"
        Header += "Peptide\tCharge\tValidProteinFlag\tFacultativeFlag\tBestSpectrum\t"
        Header += "BestModlessSpectrum\tBestModlessMQScore\tBigDBAnn\tBigDBScore\tSpectra\tModlessSpectra\tBestMQScore\t"
        Header += "BestDeltaScore\tPeptideCount\tConsensusMQScore\tPeptideLength\tCutScoreTotal\t"
        Header += "MedianCutScore\tYPresent\tBPresent\tBYIntensity\tNTT\tModdedFraction\tProteinCoverage\t"
        Header += "SpectraThisModType\tSitesThisModType\tUnmodifiedPeptideCount\tDot0.5\tShared01\tShared11\t"
        Header += "Correlation\tLogSpectrumCount\tLogPeptideLength\tLogSpecThisType\tLogSitesThisType\t"
        Header += "DeltaVsBigDB\tModelScore\tModelPValue\tSitePValue\tKnownModType\tKnownModAnnotation\t"
        Header += "KnownModScore\tKnownModSitePValue\t"
        self.OutputFile.write(Header + "\n")
        # Two more header lines, for feature-numbers and column-numbers:
        Header = "#0\t1\t2\t3\t4\t5\t6\t7\t8\t9\t10\t11\t12\t13\t14\t15\t16\t17\t18\t19\t20\t21\t22\t23\t24\t25\t26\t27\t28\t29\t30\t31\t32\t33\t34\t35\t36\t37\t38\t39\t40\t41\t42\t43\t44\t45\t46\t47\t48\t49\t"
        self.OutputFile.write(Header + "\n")
        Header = "#Feature\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t0\t1\t2\t3\t4\t5\t6\t7\t8\t9\t10\t11\t12\t13\t14\t15\t16\t17\t18\t19\t20\t21\t22\t23\t24\t25\t26\t"
        self.OutputFile.write(Header + "\n")
        ProteinRecordCount = self.DB.count("*")
        DBResidueSize = len(self.DB) - ProteinRecordCount
        self.OutputFile.write("#DatabaseSize\t%s\t\n"%DBResidueSize)
        self.OutputFile.write("#AnnotationCount\t%s\t\n"%self.AnnotationCount)
        SiteCount = len(self.PTMs.values())
        self.OutputFile.write("#SiteCount\t%s\t\n"%SiteCount)
    def ProduceConsensusSpectra(self):
        """
        We adopt a brute-force strategy: Output a consensus spectrum for each
        modified peptide species.  And, if the equivalent unmodified peptide
        was observed, then output a consensus spectrum for the unmodified
        peptide, too.
        We write a consensus CLUSTER of each modified peptide species.  
        Later on in processing, we may try MERGING two of these clusters
        (e.g. EAM+16APK, EAMA+16PK)
        """
        ClustersBuilt = {} # keep track of clusters that have ALREADY been built
        Keys = self.Peptides.keys()
        for PeptideIndex in range(len(Keys)):
            if (PeptideIndex % 100 == 0):
                print "For peptide %s/%s..."%(PeptideIndex, len(Keys))
            AnnotationKey = Keys[PeptideIndex]
            (Annotation, Charge) = AnnotationKey
            Species = self.Peptides[AnnotationKey]
            if not Species.ModifiedFlag:
                continue
            Species.ClusterPath = os.path.join(self.ConsensusClusterDir, Annotation[2], "%s.%s.cls"%(Annotation.replace("*", "-"), Charge))
            Species.ConsensusPath = os.path.join(self.ConsensusSpectrumDir, Annotation[2], "%s.%s.dta"%(Annotation.replace("*", "-"), Charge))
            ClusterContentPath = os.path.join(self.ClusterScanListDir, Annotation[2], "%s.%s.txt"%(Annotation.replace("*", "-"), Charge))
            ClusterContentFile = open(ClusterContentPath, "wb")
            #print "Creating consensus file %s"%(Species.ConsensusPath)
            #raw_input()
            Builder = BuildConsensusSpectrum.ConsensusBuilder(Species.Charge)
            MeanMQ = 0
            for Info in Species.Spectra:
                MeanMQ += Info.MQScore
            MeanMQ /= float(len(Species.Spectra))
            ValidSpectra = 0
            for Info in Species.Spectra:
                # Omit from the consensus spectra with very poor scores:
                if Info.MQScore < MeanMQ - 3.0:
                    continue
                SpectrumFilePath = self.CachedFixedFilePaths[Info.FileNameIndex]
                # Keep track of where these scans came from:
                ClusterContentFile.write("%s\t%s\t\n"%(SpectrumFilePath, Info.ByteOffset))
                Spectrum = MSSpectrum.SpectrumClass()
                SpectrumFile = open(SpectrumFilePath, "rb")
                SpectrumFile.seek(Info.ByteOffset)
                Spectrum.ReadPeaksFromFile(SpectrumFile, SpectrumFilePath)
                if not Spectrum.PrecursorMZ:
                    print "* Error: Unable to read spectrum from '%s:%s'"%(SpectrumFilePath, Info.ByteOffset)
                    continue
                ValidSpectra += 1
                Spectrum.SetCharge(Charge)
                SpectrumFile.close()
                Builder.AddSpectrum(Spectrum)
                # Special (and easy) case: If we only saw one spectrum, then write it
                # out without changing it!
                if len(Species.Spectra) == 1:
                    Spectrum.WritePeaks(Species.ConsensusPath)
            # Write the modded cluster to disk, since we may try to augment it later:
            Builder.PickleCluster(Species.ClusterPath)
            if len(Species.Spectra) > 1:
                Spectrum = Builder.ProduceConsensusSpectrum()
                Spectrum.WritePeaks(Species.ConsensusPath)
            ClusterContentFile.close()
            # If we have unmodified peptides for this species, build their cluster:
            if Species.Modless:
                Species.Modless.ConsensusPath = os.path.join(self.ConsensusSpectrumDir, Species.Modless.Annotation[2], "%s.%s.dta"%(Species.Modless.Annotation.replace("*", "-"), Charge))
                Species.Modless.ClusterPath = os.path.join(self.ConsensusClusterDir, Species.Modless.Annotation[2], "%s.%s.cls"%(Species.Modless.Annotation.replace("*", "-"), Charge))
                ModlessKey = (Species.Modless.Annotation, Species.Modless.Charge)
                if ClustersBuilt.has_key(ModlessKey):
                    pass
                else:
                    ModlessMeanMQ = 0
                    for Info in Species.Modless.Spectra:
                        ModlessMeanMQ += Info.MQScore
                    ModlessMeanMQ /= float(len(Species.Modless.Spectra))
                    Builder = BuildConsensusSpectrum.ConsensusBuilder(Species.Charge)
                    for Info in Species.Modless.Spectra:
                        # Omit from the consensus spectra with very poor scores:
                        if Info.MQScore < ModlessMeanMQ - 3.0:
                            continue
                        SpectrumFilePath = self.CachedFixedFilePaths[Info.FileNameIndex]
                        Spectrum = MSSpectrum.SpectrumClass()
                        SpectrumFile = open(SpectrumFilePath, "rb")
                        SpectrumFile.seek(Info.ByteOffset)
                        Spectrum.ReadPeaksFromFile(SpectrumFile, SpectrumFilePath)
                        Spectrum.SetCharge(Charge)
                        SpectrumFile.close()
                        Builder.AddSpectrum(Spectrum)
                    Spectrum = Builder.ProduceConsensusSpectrum()
                    Spectrum.WritePeaks(Species.Modless.ConsensusPath)
                    Builder.PickleCluster(Species.Modless.ClusterPath)
                    ClustersBuilt[ModlessKey] = 1
    def ComputeFeaturesAllPTMs(self):
        """
        Compute, and output, features for each modification site.
        """
        self.OutputFile = open(self.OutputPath, "wb")
        self.OutputPTMInfoHeader()
        # Use self.ConsensusCreatedFlags to flag which unmodified peptides we have
        # already generated consensus spectra for, so that we don't do the same one twice and waste time:
        self.ConsensusCreatedFlags = {}
        # Order the peptides by (ModDBPos, Annotation, Charge).  It's important to order things
        # in this way so that, when we combine the *large* output files for HEK293, we can keep
        # consistent 'cursors' moving through each of our input files.
        print "Sorting %s peptides..."%len(self.Peptides.values())
        SortedKeys = []
        for Peptide in self.Peptides.values():
            Key = (Peptide.ModDBPos, Peptide.Annotation, Peptide.Charge)
            SortedKeys.append(Key)
        SortedKeys.sort()
        for KeyIndex in range(len(SortedKeys)):
            (ModDBPos, Annotation, Charge) = SortedKeys[KeyIndex]
            Key = (Annotation, Charge)
            Species = self.Peptides[Key]
            if not Species.ModifiedFlag:
                continue
            if Species.DBPos < self.StartOutputDBPos:
                continue
            print "(%s/%s) PTM: %+d on db residue %d"%(KeyIndex, len(SortedKeys), Species.ModMass, Species.DBPos + Species.Peptide.Modifications.keys()[0])
            sys.stdout.flush()
            try:
                Features = self.ComputePTMFeatures(Species)
            except:
                traceback.print_exc()
                print "** Error: Unable to compute PTM features for %s"%Species
                continue
            Str = "%s\t"%self.OutputPath
            Str += "%s\t%+d\t%s\t"%(Species.ModDBPos, Species.ModMass, Species.ModAA) 
            Str += "%s\t"%Species.ProteinName
            Str += "%s\t"%(Species.ResidueNumber + Species.Peptide.Modifications.keys()[0])
            Str += "%s\t"%Species.Annotation
            Str += "%s\t"%Species.Charge
            for Feature in Features:
                Str += "%s\t"%Feature
            print Str
            self.OutputFile.write(Str + "\n")
            # We're done with this PTM now, so let's forget about it:
            del self.Peptides[Key]
        self.OutputFile.close()
    def ComputePTMFeatures(self, Species):
        """
        Compute scoring-features for this peptide species, return them as a list
        """
        Features = []
        # Feature: Is the PTM from a valid protein?  (Note: this feature is not INPUT for the
        # model, it's our desired output)
        Feature = self.GetValidProteinFlag(Species)
        Features.append(Feature)
        # Important question: Is this PTM constituitive, or facultative?  (In other words:
        # is there a spectra annotated with an UNMODIFIED peptide for this PTM type?)
        # Set flag to "1" if the PTM is facultative:
        if Species.Modless:
            Features.append("1")
        else:
            Features.append("")
        # The best spectrum observed (meta-data, not a scoring feature)
        BestMQScore = -999
        BestDeltaScore = None
        for Info in Species.Spectra:
            if Info.MQScore > BestMQScore:
                BestMQScore = Info.MQScore
                BestDeltaScore = Info.DeltaScore
                FilePath = self.CachedFixedFilePaths[Info.FileNameIndex]
                BestMQSpectrum = ("%s:%s"%(FilePath, Info.ByteOffset))
        Features.append(BestMQSpectrum)
        # The best MODLESS spectrum observed (meta-data, not a scoring feature)
        if Species.Modless:
            BestMQScoreModless = -999
            for Info in Species.Modless.Spectra:
                if Info.MQScore > BestMQScoreModless:
                    BestMQScoreModless = Info.MQScore
                    FilePath = self.CachedFixedFilePaths[Info.FileNameIndex]
                    BestMQSpectrum = ("%s:%s"%(FilePath, Info.ByteOffset))
            Features.append(BestMQSpectrum)
            Features.append(str(BestMQScoreModless))
        else:
            Features.append("")
            Features.append("")
        # Feature: Annotation, and MQScore, from a search versus big-DB.  (This feature
        # will be spiked in later)
        Features.append("")
        Features.append("")
        # Feature: Number of spectra annotated with this PTM
        Features.append(Species.SpectrumCount)
        # Feature: Number of spectra for the *unmodified* peptide version:
        if Species.Modless:
            Feature = Species.Modless.SpectrumCount
            Features.append(Feature)
        else:
            Features.append(0)
        # Feature: Best MQScore for this PTM, and the best delta-score for that scan:
        Features.append(BestMQScore)
        Features.append(BestDeltaScore)
        # Feature: Number of peptide species observed for this PTM on this residue
        Features.append(len(Species.PTM.SpeciesList))
        # Feature: Consensus annotation score (and score-features) for this peptide
        Species.ConsensusScore = None
        self.GetConsensusMQScore(Species, Features)
        # Feature: Presence of unmodified peptides covering the residue of interest
        ModlessCount = self.Coverage[Species.PTM.DBPos]
        ModdedSpectrumCount = self.ModCoverage[Species.PTM.DBPos]
        ModdedFraction = ModdedSpectrumCount / float(ModlessCount + ModdedSpectrumCount)
        Features.append(ModdedFraction)
        # Feature: Coverage of the protein of interest (ONLY FOR FACULTATIVE!)
        ProteinIndex = self.GetProteinIndex(Species.DBPos)
        ProteinCoverage = self.ProteinCoverageLevels[ProteinIndex]
        Features.append(ProteinCoverage)
        # Feature: Number of annotations using this modification-type
        ModTypeKey = (Species.ModAA, Species.ModMass)
        ModTypeSpectrumCount = self.ModTypeSpectrumCount.get(ModTypeKey, 0)
        Features.append(ModTypeSpectrumCount)
        # Feature: Number of sites using this modification-type
        ModTypeSiteCount = self.ModTypeSiteCount.get(ModTypeKey, 0)
        Features.append(ModTypeSiteCount)
        # Feature: Number of unmodified peptide species for this site
        ModlessPeptides = self.PeptideCoverage[Species.DBPos]
        Features.append(ModlessPeptides)
        # Features for FACULTATIVE PTMs only:
        # These features have been commented out, since we no longer pursue a
        # special model for facultative PTMs.
        if 0: #Species.Modless:
            Comparator = SpectralSimilarity.SpectralSimilarity(Species.ConsensusPath,
               Species.Modless.ConsensusPath, Species.Annotation, Species.Modless.Annotation)
            Comparator.LabelPeaks(0.5)
            Similarity = Comparator.DotProduct(0.5, HashByRank = 1)
            Features.append(Similarity)
            Similarity = Comparator.GetSharedPeakCount(0, 1)
            Features.append(Similarity)
            Similarity = Comparator.GetSharedPeakCount(1, 1)
            Features.append(Similarity)
            CorrelationCoefficient = Comparator.ComputeCorrelationCoefficient(1.0)
            Features.append(CorrelationCoefficient)
            del Comparator
        else:
            # This PTM is constitutive, so omit the spectrum-comparison features:
            Features.append("") # dot
            Features.append("") # shared-peaks
            Features.append("") # shared-peaks
            Features.append("") # correlation
        # Feature: Log of spectrum-count
        Features.append(math.log(1.0 + Species.SpectrumCount))
        # Feature: Log of peptide-length
        Features.append(math.log(len(Species.Peptide.Aminos)))
        # Feature: Log of same-modtype-spectrum-count
        Features.append(math.log(1.0 + ModTypeSpectrumCount))
        # Feature: Log of same-modyupe-site-count
        Features.append(math.log(1.0 + ModTypeSiteCount))
        # Feature: Delta-score versus big-db search result.  To be spiked in later!
        Features.append("")
        # Free the PySpectrum object now:
        Species.PySpectrum = None
        return Features
    def GetValidProteinFlag(self, PTM):
        # Normally we prepend "xxx" to the bogus names:
        if PTM.ProteinName[:3] == "XXX":
            return 0
        return 1
    def GetProteinIndex(self, DBPos):
        for ProteinIndex in range(len(self.ProteinPositions)):
            Pos = self.ProteinPositions[ProteinIndex]
            if Pos > DBPos:
                return ProteinIndex - 1
        return len(self.ProteinPositions) - 1
    def PreComputeAminosForMasses(self):
        """
        PepNovo often gives us partial interpretations - e.g. a
        peptide that starts at 250Da.  We "fill in" the prefix and
        suffix to generate a (not necessarily optimal) full-length
        peptide.
        """
        Aminos = "ACDEFGHILMNPQSTVWYRK" # PREFER ending in R or K.
        self.AAStrings = {}
        TotalMass = 0
        for AA1 in Aminos:
            Mass1 = Global.AminoMass[AA1]
            TotalMass = int(round(Mass1))
            self.AAStrings[TotalMass] = "%c"%(AA1)
            for AA2 in Aminos:
                Mass2 = Global.AminoMass[AA2]
                TotalMass = int(round(Mass1 + Mass2))
                self.AAStrings[TotalMass] = "%c%c"%(AA1, AA2)
                for AA3 in Aminos:
                    Mass3 = Global.AminoMass[AA3]
                    TotalMass = int(round(Mass1 + Mass2 + Mass3))
                    self.AAStrings[TotalMass] = "%c%c%c"%(AA1, AA2, AA3)
                    for AA4 in Aminos:
                        Mass4 = Global.AminoMass[AA4]
                        TotalMass = int(round(Mass1 + Mass2 + Mass3 + Mass4))
                        self.AAStrings[TotalMass] = "%c%c%c%c"%(AA1, AA2, AA3, AA4)
    def AddSpectrumToCluster(self, InputFilePath, InputFilePos, ClusterFile, Charge):
        """
        Append the specified scan to an ever-growing .mgf file
        Returns 1 if successful, 0 if failed
        """
        try:
            SpectrumFile = open(InputFilePath, "rb")
        except:
            print "** Error: couldn't open spectrum data file %s"%InputFilePath
            return 0
        SpectrumFile.seek(InputFilePos)
        Spectrum = MSSpectrum.SpectrumClass()
        try:
            Spectrum.ReadPeaksFromFile(SpectrumFile, InputFilePath)
        except:
            traceback.print_exc()
            print "***Can't parse:", InputFilePath, FileOffset
            return 0
        SpectrumFile.close()
        ParentMass = Spectrum.PrecursorMZ * Charge - (Charge - 1)*1.0078 #Peptide.Masses[-1] + 19
        #MZ = (ParentMass + (Info.Charge - 1)*1.0078) / Info.Charge
        # Now write out this spectrum to the cluster:
        self.ClusterScanNumber += 1 # ASSUMED: The caller set this to 0 at the start of the cluster!
        ClusterFile.write("BEGIN IONS\n")
        ClusterFile.write("TITLE=%s:%s\n"%(InputFilePath, InputFilePos))
        ClusterFile.write("SCAN=%s\n"%self.ClusterScanNumber)
        ClusterFile.write("CHARGE=%s\n"%Charge)
        ClusterFile.write("PEPMASS=%s\n"%ParentMass)
        for Peak in Spectrum.Peaks:
            ClusterFile.write("%s %s\n"%(Peak.Mass, Peak.Intensity))
        ClusterFile.write("END IONS\n")
        #ClusterFile.close()
        return 1
    def GetConsensusMQScore(self, Species, Features):
        """
        Feature: MQScore of the consensus spectrum.
        - Write specra to a cluster (done by ProduceConsensusSpectra)
        - Generate a consensus-spectrum for the cluster (done by ProduceConsensusSpectra)
        - Load the consensus-spectrum
        - Score the spectrum
        """
        # Load in the consensus spectrum, and score the peptide annotation:
        try:
            print ">>PyConsensus spectrum:", Species.ConsensusPath
            PySpectrum = PyInspect.Spectrum(Species.ConsensusPath, 0)
            Species.PySpectrum = PySpectrum
            print ">>ScorePeptideDetailed(%s)"%Species.Annotation
            ScoreList = PySpectrum.ScorePeptideDetailed(Species.Annotation)
            Species.ConsensusScore = ScoreList[0]
            for ScoreItem in ScoreList:
                Features.append(ScoreItem)
            print "PyInspect score %s -> %s"%(Species.Annotation, ScoreList[0])
        except:
            traceback.print_exc()
            for X in range(8):
                Features.append(0)
    def ParseCommandLine(self, Arguments):
        (Options, Args) = getopt.getopt(Arguments, "d:r:w:s:M:lp:c:Z:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-r":
                # -r results file(s)
                self.ResultsFileName = Value
            elif Option == "-c":
                self.RequiredFileNameChunk = Value
            elif Option == "-d":
                self.DBPath = Value
            elif Option == "-M":
                self.PopulateSpectrumOracle(Value)
            elif Option == "-w":
                self.OutputDir = Value
                self.OutputPath = os.path.join(self.OutputDir, "PTMFeatures.txt")
            elif Option == "-l":
                self.QuickParseFlag = 1
            elif Option == "-s" or Option == "-M":
                self.PopulateSpectrumOracle(Value)
                #self.SpectrumDir = Value
            elif Option == "-p":
                self.PValueCutoff = float(Value)
            elif Option == "-Z": # secret debugging option: Start output from DB position
                self.StartOutputDBPos = int(Value)
            else:
                print "* Error: Unrecognized option %s"%Option
            
UsageInfo = """
ComputePTMFeatures: Generate feature values for PTMs observed on a data-set.
Run this AFTER running SelectSites, and BEFORE running TrainPTMFeatures.

Arguments:
 -r [ResultsFile]: Name of the results file (or directory)
 -d [DBPath]: Path to the .trie file searched
 -w [OutputDir]: Output file directory.  Features are written to
    PTMFeatures.txt within this directory.  Clusters and other info
    is written in (or below) this directory.
 -M [RootDir]: Root directory for mzXML files. 
"""
           
if __name__ == "__main__":
    if not PROFILE_FLAG:
        try:
            import psyco
            psyco.full()
        except:
            print "(psyco not installed; running unoptimized)"
    Trainer = PTMFeatureComputer()
    Trainer.ParseCommandLine(sys.argv[1:])
    if not Trainer.ResultsFileName or not Trainer.DBPath:
        print UsageInfo
        sys.exit(-1)
    if PROFILE_FLAG:
        import profile
        profile.run("Trainer.ComputeFeaturesMain()")
    else:
        Trainer.ComputeFeaturesMain()
    
    
